	static char sccsid[]="%Z%	alloc.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	alloc.c [%P%]

Description:

	Front ends to memory allocation routines malloc(), calloc(), and realloc().
	Each routine returns 0 if zero bytes are to be allocated.
	Otherwise, if the allocation fails, each routine prints out the supplied
	error message and exits with an exit value of 2.

Functions:
	cgmalloc()
	cgmalloc2()
	cgcalloc()
	cgrealloc2()


*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include	"re.h"

/******************************************************************************

Routine:
	cgmalloc()

Syntax:

	CHAR *cgmalloc(n, s)

	int n;		// Number of bytes of memory to allocate
	char *s;	// Error message to print if allocation fails

Description:

	If n is non-zero, invokes malloc(n).
	If the allocation fails, prints out string s and exits.

Global Variables:

	None

Files:
	None

Return Status:

	Pointer to allocated memory.
	Does not return if allocation fails.
	0 if n is zero

*******************************************************************************/

CHAR *
cgmalloc(n, s)
unsigned int n;
char *s;
{
	CHAR *x;
	char buf[256];

	if (n == 0) {
		return((CHAR *)0);
	}
	if ((x = (CHAR *)malloc(n)) == 0) {
		SPR(buf, "malloc fail(%u): %s", n, s);
		perror(buf);
		exitmsg0("%s: malloc failure\n");
#ifdef DEBUG
	} else {
		printf("malloc %u for %s\n", n, s);
#endif
	}
	return(x);
}

/******************************************************************************

Routine:
	cgmalloc2()

Syntax:

	CHAR *cgmalloc2(n1, n2, s)

	int n1;		// Number of contiguous blocks of memory to allocate
	int n2;		// Size of each block in bytes
	char *s;	// Error message to print if allocation fails

Description:

	For all systems in which an "int" is the same size as a "long",
	cgmalloc2(n1, n2, s) is entirely equivalent to cgmalloc(n1*n2, s).
	This includes all supported UN*X systems.

	For systems where an "int" is smaller than a "long",  cgmalloc2()
	will exit with an error message if n1*n2 is too large to be represented
	as an integer.  Otherwise it is also entirely equivalent to
	cgmalloc(n1*n2, s).  This complication facilitates the implementation
	of cgrep for MS_DOS, because otherwise it becomes difficult to track
	down memory allocation problems in MS_DOS.

Global Variables:

	None

Files:
	None

Return Status:

	Pointer to allocated memory.
	Does not return if allocation fails.
	0 if n1 or n2 is zero

*******************************************************************************/

CHAR *
cgmalloc2(n1, n2, s)
unsigned int n1;
unsigned int n2;
char *s;
{
	unsigned int n;

	n = n1 * n2;
#ifdef __MSDOS__
	if (sizeof(unsigned int) < sizeof (unsigned long)) {
		unsigned long n_long;
		n_long = (unsigned long)n1 * (unsigned long)n2;
		if (n != n_long) {
			exitmsg2("%s: malloc %s of %lu is too big\n", s, n_long);
		}
	}
#endif
	return(cgmalloc(n, s));
}

/******************************************************************************

Routine:
	cgcalloc()

Syntax:

	CHAR *cgcalloc(n1, n2, s)

	int n1;		// Number of contiguous blocks of memory to allocate
	int n2;		// Size of each block in bytes
	char *s;	// Error message to print if allocation fails

Description:

	If n1 and n2 are non-zero, invokes calloc(n1, n2).
	If the allocation fails, prints out string s and exits.

Global Variables:

	None

Files:
	None

Return Status:

	Pointer to allocated memory.
	Does not return if allocation fails.
	0 if n1 or n2 is zero

*******************************************************************************/

CHAR *
cgcalloc(n1, n2, s)
unsigned int n1;
unsigned int n2;
char *s;
{
	CHAR *x;
	char buf[256];

	if (n1 == 0 || n2 == 0) {
		return((CHAR *)0);
	}
	if ((x = (CHAR *)calloc(n1, n2)) == 0) {
		SPR(buf, "calloc fail(%u * %u): %s", n1, n2, s);
		perror(buf);
		exitmsg0("%s: calloc failure\n");
#ifdef DEBUG
	} else {
		printf("calloc %u for %s\n", n1*n2, s);
#endif
	}
	return(x);
}

/******************************************************************************

Routine:
	cgrealloc2()

Syntax:

	CHAR *cgrealloc2(p, n1, n2, s)

	char *p;	// Pointer to previously allocated memory.
	int n1;		// Number of contiguous blocks of memory to allocate
	int n2;		// Size of each block in bytes
	char *s;	// Error message to print if reallocation fails

Description:

	For systems where an "int" is smaller than a "long",  cgrealloc2()
	will exit with an error message if n1*n2 is too large to be represented
	as an integer.  This complication facilitates the implementation
	of cgrep for MS_DOS, because otherwise it becomes difficult to track
	down memory allocation problems in MS_DOS.

	Otherwise, if n1 and n2 are non-zero, invokes realloc(p, n1*n2).
	If the reallocation fails, prints out string s and exits.

Global Variables:

	None

Files:
	None

Return Status:

	Pointer to reallocated memory.
	Does not return if reallocation fails.
	0 if n1 or n2 is zero

*******************************************************************************/

CHAR *
cgrealloc2(p, n1, n2, s)
char *p;
unsigned int n1;
unsigned int n2;
char *s;
{
	CHAR *x;
	char buf[256];
	unsigned int n;

	n = n1 * n2;
#ifdef __MSDOS__
	if (sizeof(unsigned int) < sizeof (unsigned long)) {
		unsigned long n_long;
		n_long = (unsigned long)n1 * (unsigned long)n2;
		if (n != n_long) {
			exitmsg2("%s: realloc %s of %lu is too big\n", s, n_long*16L);
		}
	}
#endif
	if ((x = (CHAR *)realloc(p, n)) == 0) {
		SPR(buf, "realloc fail(%u): %s", n, s);
		perror(buf);
		exitmsg0("%s: realloc failure\n");
#ifdef DEBUG
	} else {
		printf("realloc %u for %s\n", n, s);
#endif
	}
	return(x);
}
