	static char sccsid[]="%Z%	convert.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	convert.c [%P%]

Description:

	Converts patterns into standard format for subsequent compilation.

Functions:
	convert()
	do_convert()
	fgrep_convert()
	grep_convert()
	special()

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "re.h"

#define CHECK_OCTAL(_n) { int i; arg--;\
for (i=0; i < 3; i++) { if (*arg < '0' || *arg >= '8') break;\
(_n) <<= 3; (_n) += (*arg++ - '0'); } }

#define SPECIAL(arg, q, begin) {CHAR *a=arg; q=special(&a,q,&begin); arg = a;} 

#define EXPRESSION (char *)pat->expression

static void do_convert(), fgrep_convert(), grep_convert();
static CHAR *special();

/******************************************************************************

Routine:
	convert()

Syntax:

	void convert(pat)

	PATTERN *pat;	// Pointer to pattern to possibly be converted.

Description:

	convert() determines whether or not the pattern is to be converted
	to "full regular expression" format, which defined in the prologue
	for do_convert().

	If the conversion is to be done, convert() calls do_convert()
	to actually do it.  do_convert() will set pat->parser to egrepparse
	when the conversion is completed.

	------------------------------
	The "full regular expression" format is fully general, and the conversion
	does not affect the final compiled pattern.  However, if the pattern
	is one or more fixed strings, and there are no complicating factors,
	the conversion is unnecessary.  Therefore, for performance reasons,
	the conversion is done selectively, and the simple case of fixed
	strings is treated without conversion.

	------------------------------
	do_convert() will be called for a pattern if any of the following
	hold for that pattern:

	1) If -R (or -T) is used, and the selected pattern contains
		the application dependent string for special expansion.
	2) cgrep -F is in effect, the -a flag is in effect, and the
		pattern contains a backslash.
	3) cgrep -G mode is in effect, and the pattern contains
	     a meta-character for the mode other than NL.
	4) cgrep -E mode is in effect, and the pattern contains
	     a meta-character for the mode other than '|' or NL.
		NOTE: Until such a metacharacter is found, '|'s are converted
		to NL's by this routine.
	5) Full line or word matching is being used for this pattern (pat->xflag).
	6) Approximate matching is being used for this pattern (pat->approx).
	7) The pattern will be matched backwards. 

	------------------------------
	If none of the above (1-7) hold, the pattern is one or more fixed strings
	separated by new-lines (or by '|' that have been converted to new-lines).

	In this case, do_convert() is not needed or called, and convert()
	will do the following:

	a) If pat->parser is egrepparse, any '|' metacharacters will be
	changed in place to new-line metacharacters.  No other pattern
	conversion is done.

	b) If pat->parser is grepparse or egrepparse, pat->parser will be set
	to fgrepparse to indicate that we are in "fixed string" format.


Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

void
convert(pat)
PATTERN *pat;
{
	/* grep and egrep mode metacharacters, respectively, excluding new-line */
	char *g_special = "\\^$.[*";
	char *e_special = "\\^$.[*+?()|";
	register CHAR *s;

#ifdef	DEBUG
		printf("ORIG EXP: %s\n", EXPRESSION);
#endif

	/* If we are matching every line, or no line, no conversion is needed. */
	if (pat->parser == nlparse || pat->parser == nullparse) {
		return;
	}

	if (pat->special) {
		/* We must check for occurrence of the "special"
		 * pattern defined for the -R and -T options.
		 */
		int slen = strlen((char *)pat->special);
		register char c = (char)pat->special[0];
		register char *p = EXPRESSION;

		/* Search for the "special" pattern in the expression. */
		for (p = strchr(p, c); p; p = strchr(p+1, c)) {
			if (!strncmp(p, (char *)pat->special, slen)) {
				break;
			}
		}
		if (p) {
			/*
			 * 1) The pattern to be matched contains the application
			 * dependent string for special expansion.
			 * We must convert to "full regular expression" format.
			 */
			do_convert(pat); return;
		} else {
			/*
			 * The pattern to be matched does not contain
			 * the application dependent string for special expansion.
			 * Therefore, the special expansion is not done.
			 */
			pat->special = 0;
		}
	}

	switch (pat->parser) {
	case fgrepparse:
		if (!pat->bmap['0'] && strchr(EXPRESSION, '\\')) {
			/*
			 * 2) cgrep -F is in effect, the -a flag is in effect,
			 * and the pattern contains a backslash.
			 * We must convert to "full regular expression" format.
			 */
			do_convert(pat); return;
		}
		break;

	case grepparse:
		if (strpbrk(EXPRESSION, g_special)) {
			/*
			 * 3) cgrep -G mode is in effect, and the pattern contains
			 * a meta-character for the mode.
			 * We must convert to "full regular expression" format.
			 */
			do_convert(pat); return;
		}
		break;

	case egrepparse:
		for (s = pat->expression;
			    (s = (CHAR *)strpbrk((char *)s, e_special)) != 0; s++) {
			if (*s == '|') {
				/*
				 * cgrep -E mode is in effect, and the pattern
				 * contains the meta-character '|'.  This does
				 * not necessarily require that we convert to
				 * "full regular expression" format.
				 *
				 * We convert the '|' metacharacter to the
				 * equivalent new-line metacharacter,
				 * because '|' is not a metacharacter in
				 * either "full regular expression" or
				 * "fixed string" format.
				 */
				*s = '\n';
			} else {
				/*
				 * 4) cgrep -E mode is in effect, and the pattern
				 * contains a meta-character for the mode other
				 * than '|' or NL.
				 * We must convert to "full regular expression"
				 * format.
				 */
				do_convert(pat); return;
			}
		}
		break;
	}

	if (pat->xflag || pat->approx || pat->backward) {
		/*
		 * 5) Full line or word matching (pat->xflag), OR
		 * 6) Approximate matching (pat->approx), OR
		 * 7) The pattern will be matched backwards 
		 * We must convert to "full regular expression" format.
		 */
		do_convert(pat); return;
	}

	/* There is no need to convert to "full regular expression" format.
	 * The pattern is zero or more fixed strings, separated by new-lines.
	 * We can use "fixed string" format, and no further conversion
	 * is necessary.
	 */
	pat->parser = fgrepparse;
	return;
}

/******************************************************************************

Routine:
	do_convert()

Syntax:

	static void do_convert(pat)

	PATTERN *pat;	// Pointer to pattern to be converted.

Description:

	do_convert does the following:

	a) If applicable, do_convert() first calls fgrep_convert() or
	grep_convert() to convert cgrep -F mode or cgrep -G mode patterns
	respectively to cgrep -E mode.

	b) do_convert() then converts the cgrep -E mode pattern to "full regular
	expression" format.  A description of that format follows soon.

	c) If applicable, do_convert() implements the -x, -x1, -a, and the
	special expansion for the -R or -T operations by augmenting or modifying
	the pattern.  This is done in the same pass as the format conversion.

	d) do_convert sets pat->parser to egrepparse to indicate that the pattern
	is now in "full regular expression" format.
	-----------

	Here is a description of "full regular expression" format.   Upon
	output, the pattern will be in this format.

	"full regular expression" format is the same as that recognized by
	cgrep -E, with the following differences:

	1) '^', '$', and '|' are literals, not metacharacters.  New-line
	characters are used in place of these.  An unescaped new-line
	character (not '|') is the "OR" operator between multiple expressions,
	and an escaped new-line character (not '^' or '$') is the begin line
	or end line indicator (a literal new-line character).

	2) '<', '>', '#', and '`' are metacharacters, except within a character
	class, and must be escaped if meant literally.
	
		2a) '<' and '>' indicate that the enclosed portion of the pattern
		must be matched exactly even if approximate matching was selected.
		Angle bracket metacharacters must balance, and cannot be nested.
		Angle brackets are metacharacters even if approximate matching
		was not used.

		2b) '#' matches any single character other than a letter, number,
		or underscore.

		2c) '`' (back tic) means that backreferencing
		was used (this implies the pattern was converted from
		cgrep -G mode).  If backreferencing was used, the meaning
		of parentheses is that for grep mode backreferencing, not
		for grouping as in egrep.  For example, grep mode
		"...\(...\)...\1..." is converted to "...(...)..`1..."

		In, contrast, in cgrep -E format, '<' and '>' are metacharacters
		only if approximate matching is used, and they may be nested.
		'#' and '`' are metacharacters only if the pattern was converted
		from cgrep -G format.

	3) '\' (backslash) is a metacharacter even inside a character class.
	It means that the next character is a literal.  For example, to
	represent a literal backslash inside of a character class, two
	consecutive backslashes are required.
	In, contrast, in cgrep -E format, a backslash in a character class
	is a metacharacter only if the -a option was used.
	-----------

	Converting to "full regular expression" format from cgrep -E format
	involves copying the cgrep -E mode pattern as is, except for the
	following modifications:

	1) If pat->special is set, call special() to do the special expansion
	due to the (-R or -T flags).

	2) If approximate matching is used, and angle brackets are nested,
	all but the outermost are angle brackets are discarded.

	3) If -x or -x1 was used, enhance the pattern to
	force it to match an entire line (-x) or a complete word or
	phrase (-x1).  The enhanced portions will be surrounded by
	angle brackets so that, even if approximate matching is in effect,
	they must be matched exactly.

	4) Null expressions are ignored, to avoid fatal compiler errors.
	This is being user friendly.

	5) All metacharacters in cgrep -E mode that are not metacharacters in
	"full regular expression" format must be converted to equivalents.

	These are:
	'^' is converted to backslash newline (a literal new line).
	'$' is converted to backslash newline (a literal new line).
	'|' is converted to newline.

	6) If -a was used, convert backslashed expressions to their values,
	(e.g. \n to \<new-line>).

	7) All literal characters in cgrep -E mode that are metacharacters in
	"full regular expression" format must be escaped with a backslash.

	These are:
	'<' and '>', if approximate matching is not used.
	'\' itself, if inside a character class and -a was not used.
	'`', if not backreferencing.
	'#', unless we converted from cgrep -G mode. 

	8) If we converted from cgrep -G mode, some characters need additional
	handling.

	These are:
	'(' and ')':  These have been converted from '\(' and '\)'.  If we are
	not backreferencing, these have no effect, and are discarded.
	'#': This has been converted from '\<' or '\>'.  Surround '#' with angle
	brackets to ensure an exact match, unless it is already inside angle
	brackets.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
do_convert(pat)
PATTERN *pat;
{
	register CHAR *arg;
	register CHAR *q;
	register BOOL backslash;
	register int  paren_cnt = 0;
	register int  angle_cnt = 0;
	BOOL begin = YES;
	CHAR	prevchar = '\0';
	long buflen;

	switch (pat->parser) {
	case grepparse:
		/* grep mode was used for this pattern.
		 * Convert the pattern to its equivalent pattern in
		 * egrep mode.  (NOTE: Backreferencing has no egrep
		 * equivalent, so we have defined a new metacharacter for it).
		 */
		grep_convert(pat);
		break;

	case fgrepparse:
		/* fgrep mode was used for this pattern.
		 * Convert the pattern to its equivalent pattern in
		 * egrep mode.
		 */
		fgrep_convert(pat);
		break;
	}
#ifdef	DEBUG
	printf("OLD EXP: %s\n", EXPRESSION);
#endif
	/* Now the pattern is in egrep format.  Convert it to
	 * the "full regular expression" internal format.
	 */
	arg = pat->expression;			

	/* Allocate a buffer large enough to hold the converted string. */
	buflen = 2 * strlen(EXPRESSION) * sizeof(CHAR);
	if (pat->xflag) buflen *= 2;
	buflen += ESIZE;
	q = (CHAR *)cgmalloc((unsigned)buflen, "do_convert");

	pat->expression = q;

	--arg;
	/* Now loop through the characters */
	while (1) {
		arg++;
		/* 1: If the special pattern for the -R (or -T) option is
		 * present, and we are at the beginning of an expression
		 * outside of any parentheses, expand any special
		 * pattern that may begin this expression.
		 */
		if (begin && !paren_cnt && pat->special) {
			SPECIAL(arg, q, begin);
		}
		/* 2: If we are approximate matching, check for angle
		 * bracket metacharacters.  We output only the outermost
		 * pairs, because nested angle brackets are redundant.
		 */
		if (pat->approx) {
			switch (*arg) {
			case '<':
				if (angle_cnt++ == 0) {
					*q++ = *arg;
				}
				continue; /* with while loop */

			case '>':
				if (--angle_cnt == 0) {
					*q++ = *arg;
				}
				if (angle_cnt < 0) {
					re_error("Unbalanced <>.");
				}
				continue; /* with while loop */
			}
		}
		/* 3: If -x or -x1 was used, check if we need to expand the
		 * option here.  We do so only if we are at the beginning
		 * or end of an outermost expression (i.e. outside any
		 * parentheses).
		 */
		if (pat->xflag && !paren_cnt) {
			/* If we are beginning an expression, we must expand
			 * the option here, unless the first character of the
			 * expression is a beginning of line ('^'), in which
			 * case, no expansion is desired.
			 */
			if (begin && *arg != '^') {
				/* To ensure exact match, put inside
				 * angle brackets, unless already inside.
				 */
				if (!angle_cnt) {
					*q++ = '<';
				}
				if (pat->xflag == 2) {
					/* -x1 option. Insert metacharacter
					 * '#', which matches letters,
					 * numbers, and underscore.
					 */
					*q++ = '#';
				} else {
					/* -x option. Insert literal
					 * (i.e., escaped) new-line.
					 */
					*q++ = '\\';
					*q++ = NL;
				}
				if (!angle_cnt) {
					*q++ = '>';
				}
				/* We are no longer at beginning of expression. */
				begin = NO;
			}
			/* If we are ending an expression, we must expand
			 * the option here, unless the final character of
			 * the expression was an end of line ('$'), in which
			 * case no expansion is desired.
			 */
			if (prevchar != '$') {
				switch (*arg) {
				case '\n':
				case '|':
				case '\0':
					/* Alternation metacharacter '\n' or '|',
					 * or end of entire pattern.
					 */
					/* To ensure exact match, put inside
					 * angle brackets, unless already inside.
					 */
					if (!angle_cnt) {
						*q++ = '<';
					}
					if (pat->xflag == 2) {
						/* -x1 option. Insert metacharacter
						 * '#', which matches letters,
						 * numbers, and underscore.
						 */
						*q++ = '#';
					} else {
						/* -x option. Insert literal
						 * (i.e., escaped) new-line.
						 */
						*q++ = '\\';
						*q++ = NL;
					}
					if (!angle_cnt) {
						*q++ = '>';
					}
					/* We are not at beginning of expression. */
					begin = NO;
				}

			}
		}

		/* If we are done with the pattern, break out of loop. */
		if (*arg == '\0') {
			if (begin) {
				/* If the last character output was a new-line
				 * metacharacter, back up over it.  To be
				 * user friendly, we don't output it, because
				 * the resulting null expression will cause the
				 * compilation to fail.
				 */
				begin = NO;
				q--;
			}
			break;
		}

		/* Save the character */
		prevchar = (*arg);

		/* Now process the character. */
		switch (*arg) {
		case '\n':
		case '|':
			/* 4: 5: We have an alternation metacharacter. Output 
			 * a new-line alternation metacharacter.  However,
			 * to be user friendly, we don't output it if we are
			 * at the beginning of an expression, because
			 * the resulting null expression will cause the
			 * compilation to fail.
			 */
			if (!begin) {
				*q++ = '\n';
				/* We are at the beginning
				 * of the next expression.
				 */
				begin = YES;
			}
			continue; /* with while loop */
		case '^':
		case '$':
			/* 5: Begin or end line metacharacter. Convert it
			 * to a literal new-line.  Surround it by
			 * angle brackets to guarantee an exact match,
			 * unless we are already inside angle brackets.
			 */
			if (!angle_cnt) {
				*q++ = '<';
			}
			*q++ = '\\';
			*q++ = NL;
			if (!angle_cnt) {
				*q++ = '>';
			}
			break;

		case '\\':
			/* Backslash metacharacter.  Store it,
			 * and advance to next character.
			 */
			*q++ = *arg++;
			/* 6: Store the next character using the backslash
			 * mapping (which handles possible -a option).
			 * If -a was not used, "bmap" is the identity mapping.
			 */
			if (!*arg) {
				re_error("Backslash metacharacter at end of pattern.");
			}
			*q = pat->bmap[*arg];
			/* If we mapped to null, this means that we have an
			 * octal representation of an ASCII character.  Handle it.
			 */
			if (!*q) {
				arg++;
				CHECK_OCTAL(*q);
				--arg;
			}
			/* Advance past stored character. */
			q++;
			break;

		case '[':
			/* Character class metacharacter.  Store it,
			 * and advance to next character.
			 */
			*q++ = *arg++;
			/* Store initial '^', ']', or "^]",
			 * and advance to next character.
			 * An initial ']' or "^]" does not terminate
			 * the character class.
			 */
			if (*arg == '^') {
				*q++ = *arg++;
			}
			if (*arg == ']') {
				*q++ = *arg++;
			}
			/* 6: 7: We now step through the remainder of the character
			 * class, storing each character as we go.  However,
			 * if -a was used, backslashes are metacharacters,
			 * even inside a character class, so they need special
			 * processing.
			 */
			backslash = NO;
			while (*arg != ']' && *arg) {
				if (*arg == '\\') {
					/* Backslash character. if -a was not used,
					 * it is a literal backslash, and it
					 * must be escaped.
					 * If -a was used, it is a metacharacter,
					 * and subsequent characters may be
					 * affected, so we set "backslash".
					 */
					if (pat->bmap['0']) {
						*q++ = '\\';
					} else {
						backslash = YES;
					}
					/* Store backslash, and advance
					 * to next character.
					 */
					*q++ = *arg++;
				} else if (backslash) {
					/* The previous character was a backslash
					 * metacharacter.  Store the next character
					 * using the backslash mapping
					 * (which handles -a option).
					 */
					*q = pat->bmap[*arg++];
					/* If we mapped to null, this means that
					 * we have an octal representation of
					 * an ASCII character.  Handle it.
					 */
					if (!*q) {
						CHECK_OCTAL(*q);
					}
					/* Advance past stored character. */
					q++;
					backslash = NO;
				} else {
					/* Store character, and advance
					 * to next character.
					 */
					*q++ = *arg++;
				}
			}
			if (!*arg) {
				re_error("Unterminated character class.");
			}
			/* Store the terminating ']'. */
			*q++ = *arg;
			break;

		case '(':
		case ')':
			/* 8: If we converted from cgrep -G mode, this has
			 * been converted from \( or \).  Unless we are
			 * backreferencing, we discard it, because it has no
			 * effect.
			 */
			if (pat->parser == grepparse) {
				if (pat->backref) {
					*q++ = *arg;
				}
				break;
			}
			/* This is a metacharacter that is unchanged in meaning
			 * from cgrep -E mode  to "full regular expression" mode.
			 * Store it as is, but check for balance.
			 */
			if (*arg == '(') {
				paren_cnt++;
			} else if (--paren_cnt < 0) {
				re_error("Unbalanced () for grouping.");
			}
			*q++ = *arg;
			break;

		case '`':
			/* 7: If we are backreferencing, store this
			 * backreferencing metacharacter and the following
			 * character as is.  If we are not backreferencing,
			 * this is a literal character that must be escaped.
			 */
			if (pat->backref) {
				*q++ = *arg++;
			} else {
				*q++ = '\\';
			}
			*q++ = *arg;
			break;

		case '#':
			/* 7: 8: If we converted from cgrep -G mode, this is
			 * a begin or end word metacharacter, which we pass
			 * through as is, except we make sure it is inside
			 * angle brackets.
			 */
			if (pat->parser == grepparse) {
				if (!angle_cnt) {
					*q++ = '<';
				}
				*q++ = *arg;
				if (!angle_cnt) {
					*q++ = '>';
				}
				break;
			}
			/* This is a literal character that must
			 * be escaped.
			 */
			*q++ = '\\';
			*q++ = *arg;
			break;
		case '<':
		case '>':
			/* 7: A literal character that must be escaped.
			 * (We are not using approximate matching).
			 */
			*q++ = '\\';
			*q++ = *arg;
			break;

		case '+':
		case '?':
		case '*':
		case '.':
		default:
			/* A literal character, or a metacharacter that
			 * is unchanged in meaning from cgrep -E mode
			 * to "full regular expression" mode.  Store
			 * it as is.
			 */
			*q++ = *arg;
			break;

		}
		/* We are no longer at the beginning of an expression. */
		begin = NO;
	}
	/* Check for unbalanced parentheses or angle brackets. */
	if (paren_cnt) {
		re_error("Unbalanced () for grouping.");
	}
	if (angle_cnt) {
		re_error("Unbalanced <>.");
	}
	if (q >= pat->expression + buflen) {
		re_error("conversion overflow");
	}
	/* Set pointer to end of converted expression. */
	pat->exp_end = q;
	/* The converted expression needs to be null terminated. */
	*q = '\0';
	/* We have finished converting to "full regular expression" format. */
	pat->parser = egrepparse;
#ifdef	DEBUG
	printf("NEW EXP: %s\n", EXPRESSION);
#endif
	/* If the converted expression is zero length, the compilation
	 * will fail with an error.   In this case, we wish to match
	 * no lines.  So change the parser to nullparse, to avoid the error.
	 */
	if (strlen(EXPRESSION) == 0) {
		pat->parser = nullparse;
	}
}

/******************************************************************************

Routine:
	fgrep_convert()

Syntax:

	static void fgrep_convert(pat)

	PATTERN *pat;	// Pointer to pattern to be converted.

Description:

	fgrep_convert() is called by do_convert() if the pattern to
	be converted is in cgrep -F mode.  fgrep_convert() converts
	the expression from cgrep -F mode to cgrep -E mode by escaping
	all the characters that have special meaning in cgrep -E
	mode, but are taken literally in cgrep -F mode.  These are:
	'[', '^', '$', '(', ')', '*', '+', '?', '|', '.', and, if
	the -a option was not specified, '\' (backslash).

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
fgrep_convert(pat)
PATTERN *pat;
{
	CHAR *arg = pat->expression;			
	CHAR *q;
	long buflen;

	/* Allocate a buffer large enough to hold the converted string. */
	buflen = ((3*(long)strlen(EXPRESSION))/2 + ESIZE)*sizeof(CHAR);
	q = (CHAR *)cgmalloc((unsigned)buflen, "fgrep_convert");

	pat->expression = q;
	/* Loop through the pattern's characters. */
	while (*arg) {
		switch (*arg) {
		case '\\':
			/*
			 * Backslash is not escaped if the -a flag is
			 * used, because, if -a is used, backslash
			 * has same meaning in either mode.
			 */
			if (!pat->bmap['0']) {
				/* The -a flag was used.  Backslash is a
				 * metacharacter in both fgrep and egrep mode,
				 * so it should not be escaped.  Also, the
				 * following character should also not be
				 * escaped, since it is already preceded by
				 * a backslash.
				 */
				*q++ = *arg++;
				if (*arg) {
					*q++ = *arg++;
				}
				continue;
			}
			/* -a was not used.  The backslash is a literal
			 * and must be escaped.
			 */
			/* FALLTHROUGH */
		case '[':
		case '^':
		case '$':
		case '(':
		case ')':
		case '*':
		case '+':
		case '?':
		case '|':
		case '.':
			/* The character is a cgrep -E metacharacter,
			 * and is a literal in cgrep -F.  We need to escape it
			 * so it will be interpreted as a literal.
			 */
			*q++ = '\\';
			break;
		case '\n':
		default:
			/* The character has the same meaning in cgrep -F mode
			 * as in cgrep -E mode.  We need not escape it.
			 */
			break;
		}
		/* Copy the character */
		*q++ = *arg++;
	}
	if (q >= pat->expression + buflen) {
		re_error("-F mode conversion overflow");
	}
	/* The converted expression needs to be null terminated. */
	*q = '\0';
}

/*
 * grep_convert() is called by do_convert() if the pattern to
 * be converted is in cgrep -G mode.
 * grep_convert() converts the cgrep -G pattern to the equivalent
 * cgrep -E pattern.
 * In the backreferencing case, there is no cgrep -E equivalent, so
 * we use '(' for '\(', ) for '\)', and '`' for '\'.
 * <TO BE COMMENTED>
 */
/******************************************************************************

Routine:
	grep_convert()

Syntax:

	static void grep_convert(pat)

	PATTERN *pat;	// Pointer to pattern to be converted.

Description:

	grep_convert() is called by do_convert() if the pattern to
	be converted is in cgrep -G mode.  grep_convert() converts
	the expression from cgrep -G mode to cgrep -E mode.

	This involves the following possible changes to the pattern:

	1) '(', ')', '+', '?', and '|' are literals in cgrep -G mode
	but metacharacters in cgrep -E mode.  They must be escaped when
	they occur outside of a character class.

	2) '^', '$', and '*' are sometimes literals and sometimes
	metacharacters in cgrep -G mode but are metacharacters in
	cgrep -E mode.  It must be determined if each occurrence of
	'^', '$', or '*' is a literal or a metacharacter. It must
	be escaped if it occurs as a literal outside of a character
	class.  In addition, '*' is sometimes ignored in cgrep -G mode,
	in which case it is dropped.

	3) '\{' is sometimes a literal '{' in cgrep -G mode, and sometimes
	the beginning of a repeat range count.  If it is the beginning of
	a repeat range count, the repeat range count and the character or
	character class that it affects will be replaced by a cgrep -E
	equivalent.  Examples are:

	cgrep -G		cgrep -E equivalent

	a\{3\}			aaa
	a\{3,5\}		aaaa?a?
	a\{3,\}			aaa+
	a\{0,4\}		a?a?a?a?
	a\{0,\}			a*

	4) Backreferencing has no equivalent in cgrep -E mode.  Therefore
	we define a new metacharacter '`' (back tic).
	'...\(...\)...\digit...' is converted to '...(...)...`digit...'.

	5) '\<' and '\>' have no equivalent in cgrep -E mode.  Therefore
	we define a new metacharacter '#'.  '\<' or '\>' metacharacters
	are each converted to '#'.

	6) Literal '`' or '#' characters must be escaped when they occur
	outside of a character class, in order to distinguish them from
	the new metacharacters defined above.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
grep_convert(pat)
PATTERN *pat;
{
	register CHAR *arg;
	register CHAR *q;
	CHAR *starrable_ptr = 0;
	CHAR *star_ptr = 0;
	CHAR *kill_ptr = 0;
	int backrefs = 0;
	int paren_cnt = 0;
	int i, m, n, size;
	int *m_or_n;
	int angles = 0;
	BOOL range;
	BOOL begin = YES;
	long buflen;
	enum {
		D_NONE,
		D_LINE,
		D_WORD
	} end = D_NONE;

	arg = pat->expression;			

#ifdef	DEBUG
	printf("OLD -G EXP: %s\n", EXPRESSION);
#endif
	/* Allocate a buffer large enough to hold the converted string. */
	buflen = ((3*(long)strlen(EXPRESSION))/2 + ESIZE)*sizeof(CHAR);
	q = (CHAR *)cgmalloc((unsigned)buflen, "grep_convert");
	pat->expression = q;

	/* Loop through the characters. */
	/* Note that the following parameters may be set based on the identity
	 * of each character.  We explain these parameters here, rather than
	 * duplicating the same comments extensively.
	 *
	 * Angle bracket metacharacters (used in approximate matching)
	 * do not affect any of these parameters.
	 *
	 * BOOL begin;	// Are we at the beginning of an expression?
	 *		//
	 * 		// Set to YES initially, and after an unescaped
	 * 		//   new-line (expression separator).  Note that
	 * 		//   cgrep -G takes multiple expressions, even though
	 * 		//   grep does not.
	 * 		// Unaffected by angle bracket metacharacters.
	 * 		// Set to NO after any other character.
	 * 		//
	 * 		// This parameter is needed to know how to interpret
	 * 		//   '^' or '\<'.
	 *
	 * CHAR *starrable_ptr;
	 * 		// Does the character (or character sequence (e.g. a
	 * 		//   character class)) accept a following '*' character
	 * 		//   to mean zero or more repetitions?
	 *		//
	 * 		// If so, the ptr is set to the beginning of the character
	 * 		//   sequence.  This is done for any literal, for the
	 * 		//   '.' metacharacter, for a character class [...],
	 * 		//   or for the '\digit' used in backreferencing.
	 * 		// The ptr is unaffected by angle bracket metacharacters,
	 * 		//   or by the '*' metacharacter itself.
	 * 		// Otherwise, the pointer will be 0.  This occurs
	 * 		//   for the metacharacters '^', '$', '\<', '\>',
	 * 		//   '\(', '\)', '\{...\}', new-line, and initially.
	 * 		//
	 * 		// This parameter is needed to know how to interpret '*'.
	 *
	 * CHAR *star_ptr;
	 *		// Are we at a '*' metacharacter?
	 *		//
	 *		// Set to point to a '*' metacharacter, when one is
	 *		//   encountered, unless it was already non-zero,
	 *		//   in which case it is unaffected by subsequent
	 *		//   '*' metacharacters.
	 * 		// Unaffected by angle bracket metacharacters.
	 * 		// Set to 0 after any other character, and initially.
	 * 		//
	 * 		// This parameter is needed to know how to interpret '*'.
	 *
	 * CHAR *kill_ptr;
	 *		// Are we at a '\(', '\)' or '\{...\}' metacharacter?
	 * 		//
	 *		// Set to point to the beginning of the metacharacter
	 *		//   sequence, when one of the above metacharacters
	 *		//   is encountered, unless it was already non-zero,
	 *		//   in which case it is unaffected by '\{...\}.
	 * 		// Unaffected by angle bracket metacharacters.
	 * 		// Set to 0 after any other character, and initially.
	 * 		//
	 * 		// This parameter is needed to know how to interpret '*'
	 * 		//   and '\{'.
	 */
	--arg;
	while (1) {
		arg++;

		/* We temporarily defer processing of angle bracket metacharacters
		 * (used in approximate matching).  This is because we may
		 * need to determine if we are at the end of the expression
		 * in order to properly interpret '$' or '\>'.  Angle bracket
		 * metacharacters are ignored when it comes to determine
		 * whether or not we are at the beginning or end of an expression.
		 * We count the angle brackets so we can process them later.
		 */
		if (pat->approx) {
			switch (*arg) {
			case '<':
			case '>':
				angles++;
				continue; /* with while loop */
			}
		}

		if (end == D_LINE) {
			/* The previous character was an (unescaped) '$'.
			 * We deferred processing it until now.
			 * It is either a literal, in which case it must
			 * be escaped, or an end of line metacharacter, in
			 * which case it is stored as is.
			 */
			starrable_ptr = star_ptr = kill_ptr = 0;
			if (*arg && *arg != '\n') {
				/* The '$' is not at the end of an expression.
				 * It is a literal, and must be escaped.
				 */
				starrable_ptr = q;
				*q++ = '\\';
			}
			*q++ = '$';
		} else if (end == D_WORD) {
			/* The previous characters were '\>'.
			 * We deferred processing it until now.
			 * It is either a literal, in which case it must
			 * be stored as is (it is already escaped),
			 * or an end of word metacharacter, in which case it
			 * must be converted to '#'.
			 */
			starrable_ptr = star_ptr = kill_ptr = 0;
			if (*arg && *arg != '\n') {
				/* The '\>' is not at the end of an expression.
				 * It is a literal, and must be stored as such.
				 */
				starrable_ptr = q;
				*q++ = '\\';
				*q++ = '>';
			} else if (!pat->xflag) {
				/* The '\>' is at the end of an expression.
				 * It is an end of word metacharacter,
				 * and must be stored as '#'.  However,
				 * we store it only if -x or -x1 is not in effect.
				 */
				*q++ = '#';
			}
		}
		/* We have processed any deferred '$' or '\>'. */
		end = D_NONE;

		/* Now that we have handled any possible '$' or '\>' from a
		 * previous pass, we can store any angle bracket metacharacters
		 * that we may have passed over.
		 */
		for (i = 0; i < angles; i++) {
			*q++ = *((arg - angles) + i);
		}
			
		/* We have processed any deferred angle bracket metacharacters. */
		angles = 0;

		/* If we have no more characters in pattern, break out of loop. */
		if (!*arg) {
			break; /* out of while loop */
		}

		/* Loop through the characters. */
		switch (*arg) {
		case '\\':
			/* We have a backslash.  Get the next character. */
			switch (*(++arg)) {
			case '<':
				/* We have '\<'.
				 * It is either a literal, in which case it must
				 * be stored as is (it is already escaped),
				 * or a beginning of word metacharacter, in which
				 * case it must be converted to '#'.
				 */
				starrable_ptr = star_ptr = kill_ptr = 0;
				if (!begin) {
					/* The '\>' is not at the beginning of an
					 * expression.  It is a literal, and must
					 * be stored as such.
					 */
					starrable_ptr = q;
					*q++ = '\\';
					*q++ = '<';
				} else if (!pat->xflag) {
					/* The '\<' is at the beginning of an
					 * expression.
					 * It is an end of word metacharacter,
					 * and must be stored as '#'.  However,
					 * we store it only if -x or -x1 is not
					 * in effect.
					 */
					*q++ = '#';
				}
				begin = NO;
				break;

			case '>':
				/* We have '\>'.
				 * This is an end of word metacharacter if it
				 * is at the end of an expression, but is a literal
				 * otherwise.  Since we don't yet know which, we just
				 * set "end" to D_WORD to remember the this.
				 * We store nothing at this point.
				 */
				end = D_WORD;
				starrable_ptr = star_ptr = kill_ptr = 0;
				begin = NO;
				break;

			case '{':
				/* We have '\{'.
				 * In cgrep -G mode, as in grep, '{' may be a
				 * metacharacter (the beginning of a repeat range
				 * count) or a literal.
				 */
				begin = NO;
				if (!starrable_ptr && !kill_ptr) {
					/* '\{' appears either at the beginning
					 * of an expression, or after a
					 * metacharacter that does not accept
					 * a '*' metacharacter. 
					 *
					 * Furthermore, '\{' does NOT appear
					 * immediately after another repeat
					 * range count (e.g. \{m,n\}), or
					 * immediately after '\(' or '\)'.
					 *
					 * Therefore, '\{' will be interpreted
					 * as a literal.  Store it, but there
					 * is no need to store the backslash,
					 * since '{' is not a metacharacter
					 * in cgrep -E mode.
					 */
					starrable_ptr = q;
					star_ptr = 0;
					*q++ = *arg;
					break;

				}
				/* It is the beginning of a repeat range count. */

				/* If the preceding character was a '*'
				 * metacharacter, back up over it to discard it.
				 * This may be silly, but that's the way grep works.
				 */
				if (star_ptr) {
					q = star_ptr;
				}
				star_ptr = 0;

				/* Now, process the repeat range count. */
				/* Set up pointer to minimum count */
				m_or_n = &m;
				/* Initialize counts to -1, i.e. unspecified. */
				m = n = -1;
				/* Loop over the range count specification. */
				arg++;
				while (*arg != '\\') {
					switch (*arg) {
					case '0':
					case '1':
					case '2':
					case '3':
					case '4':
					case '5':
					case '6':
					case '7':
					case '8':
					case '9':
						/* We have a digit. */
						if (*m_or_n > 0) {
							/* Multiply the
							 * previous value of the
							 * count by 10, and add the
							 * value of the digit.
							 */
							*m_or_n *= 10;
							*m_or_n += (*arg - '0');
						} else {
							/* This is the first
							 * (non-zero) digit.
							 * Store its value.
							 */
							*m_or_n = (*arg - '0');
						}
						break;
						
					case ',':
						/* We have a comma.  This is
						 * the separator between the
						 * minimum and maximum counts.
						 * Set the pointer to point
						 * to the maximum count.
						 * It's an error if it already
						 * does.
						 */
						if (m_or_n == &n) {
							re_error("Bad number.");
						}
						m_or_n = &n;
						break;
					default:
						re_error("Bad number.");
					}
					arg++;
				}
				if (m < 0) {
					re_error("Range minimum not specified.");
				}
				/* If no comma separator was used, the maximum
				 * count is just the minimum count.
				 */
				if (m_or_n == &m) {
					n = m;
				}
				/* The maximum count allowed by grep is 256. */
				if (m >= 256 || n >= 256) {
					re_error("Range endpoint too large.");
				}
				/* The minimum count cannot exceed the maximum
				 * count.  NOTE: A maximum count of -1
				 * really means no maximum was specified,
				 * so don't give an error in that case.
				 */
				if (n >= 0 && m > n) {
					re_error("First number exceeds second in \\{ \\}.");
				}
				/* We stopped at a backslash.  Make sure that
				 * the range count is properly terminated by
				 * a closing brace.
				 */
				if (*++arg != '}') {
					re_error("Bad number.");
				}
				/* We have read in the entire range count. */
				if (kill_ptr) {
					/* '\{' appears immediately after another
					 * range count (e.g. \{m,n\}), or
					 * immediately after '\(' or '\)'.
					 * Ignore it.  This may be silly, but
					 * that's the way grep works.
					 */
					break;
				}
				/* Now, we must convert the range counts m and
				 * n to cgrep -E format.  The only repeat
				 * counts recognized by cgrep -E are the
				 * '*', '+', and '?' metacharacters.
				 */
				/* Set kill_ptr to point to our current
				 * storage location.
				 */
				kill_ptr = q;
				range = YES;
				/* "starrable_ptr" points to the beginning of
				 * the portion of the stored expression to which
				 * the repeat count applies.  So, calculate
				 * the length of the string to which the
				 * repeat count applies.
				 */
				size = kill_ptr - starrable_ptr;
				/* Back up over the copy of the string already
				 * stored.
				 */
				q -= size;
				/* Store 'm' copies of the string sequentially.
				 * 'm' is the minimum number of copies that
				 * need to be matched.
				 */
				for (i = 0; i < m; i++) {
					(void)strncpy((char *)q,
					  (char *)starrable_ptr, size);
					q += size;
				}
				/* See if there is a maximum number of repeats. */
				if (n >= 0) {
					/* There is a maximum number, 'n'.  Store
					 * 'n - m' copies of the string, with
					 * a '?' (meaning zero or one occurrence)
					 * after each copy.
					 */
					for (i = m; i < n; i++) {
						(void)strncpy((char *)q,
						  (char *)starrable_ptr, size);
						q += size;
						*q++ = '?';
					}
				} else {
					/* There is no upper limit, as a comma
					 * separator was used (i.e. 'm_or_n'
					 * points to 'n'), but no number was
					 * specified (e.g. \{3,\}).
					 */
					if (m) {
						/* At least 'm' occurrences are
						 * required for a match.  Add
						 * a '+' metacharacter (meaning
						 * one or more occurrences)
						 * after the last occurrence.
						 */
						*q++ = '+';
					} else {
						/* Zero or more occurrences
						 * are required for a match.  Add
						 * a '*' metacharacter (meaning
						 * zero or more occurrences)
						 * after the one occurrence
						 * previously stored.  Since
						 * we backed up over it, we
						 * must advance over it first.
						 */
						q += size;
						*q++ = '*';
					}
				}
				break;

			case ')':
				/* We have '\)'.  This is the end of a
				 * grouping for backreferencing.  Store it
				 * as ')', i.e., do not store the backslash.
				 * Check for unbalanced parentheses.
				 * Increment the count of legal backreferences.
				 */
				begin = NO;
				if (--paren_cnt < 0) {
					re_error("\\( \\) imbalance");
				}
				backrefs++;
				starrable_ptr = star_ptr = 0;
				kill_ptr = q;
				range = NO;
				*q++ = *arg;
				break;

			case '(':
				/* We have '\('.  This is the beginning of a
				 * grouping for backreferencing.  Store it
				 * as '(', i.e., do not store the backslash.
				 * Keep a count to check for parentheses
				 * imbalance.
				 */
				begin = NO;
				paren_cnt++;
				starrable_ptr = star_ptr = 0;
				kill_ptr = q;
				range = NO;
				*q++ = *arg;
				break;

			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				/* We have backslash followed by a digit from 1-9. */
				/* Is this a backreference attempt, or
				 * a literal ASCII character code (-a flag)?
				 */
				begin = NO;
				if ((int)(*arg - '0') <= backrefs) {
					/* This could be a legal backreference.
					 * Assume it is a backreference.
					 */
					pat->backref = YES;
					if (pat->approx) {
						exitmsg0("%s: backreferencing and approximation are not compatible.\n");
					}
					starrable_ptr = q;
					star_ptr = kill_ptr = 0;
					*q++ = '`';
					*q++ = *arg;
					break;
				}
				/* Not a legal backreference.  If -a flag was
				 * not used, give an error.  If -a was used,
				 * interpret the digit as a literal by
				 * falling into the default case.
				 */
				if (pat->bmap['0']) {
					re_error("\\digit out of range.");
				}
				/* FALLTHROUGH */
			default:
				/* We have backslash followed by a literal.
				 * Store the backslash and literal as is.
				 */
				begin = NO;
				starrable_ptr = q;
				star_ptr = kill_ptr = 0;
				*q++ = '\\';
				*q++ = *arg;
				break;
			}
			break;

		case '[':
			/* We have a character class.  The entire character
			 * class will be stored as is, since it means the same
			 * in cgrep -E mode as in cgrep -G mode.
			 */
			begin = NO;
			starrable_ptr = q;
			star_ptr = kill_ptr = 0;
			/* Store the '[', and advance to the next character.  */
			*q++ = *arg++;
			/* A leading ']' or '^]' does not terminate the class.  */
			if (*arg == '^') {
				*q++ = *arg++;
			}
			if (*arg == ']') {
				*q++ = *arg++;
			}
			/* Store the remainder of the class, up to the ']'.
			 * However, if -a was used, a '\]' does not terminate
			 * the class, as the ']' is then a literal.
			 */
			if (pat->bmap['0']) {
				/* -a was not used. */
				while (*arg != ']' && *arg) *q++ = *arg++;
			} else {
				while (*arg != ']' && *arg) {
					/* -a was used. If we encounter
					 * a backslash, save the next
					 * character also, without checking if
					 * it is a ']'.
					 */
					if (*arg == '\\') {
						*q++ = *arg++;
						if (*arg) *q++ = *arg++;
					} else {
						*q++ = *arg++;
					}
				}
			}
			if (!*arg) {
				re_error("Unterminated character class.");
			}
			/* Now, store the ']'. */
			*q++ = *arg;
			break;

		case '*':
			/* In cgrep -G mode, as in grep, '*' may be a
			 * metacharacter, a literal, or ignored.
			 */
			/* If '*' appears immediately after a repeat range
			 * count (e.g. \{m,n\}), it is ignored.
			 * This may be silly, but that's the way grep works.
			 */
			if (kill_ptr && range) {
				break;
			}
			/* If '*' appears immediately after another star
			 * metacharacter (not a literal '*'), it can be ignored.
			 */ 
			if (star_ptr) {
				break;
			}
			/* It's a literal or a metacharacter. */
			if (!starrable_ptr) {
				/* '*' is a literal, since either it is at
				 * the beginning of an expression, or the
				 * preceding character is a metacharacter
				 * that does not accept a '*'.
				 * Therefore, '*' must be escaped.
				 */
				starrable_ptr = q;
				star_ptr = 0;
				*q++ = '\\';
			} else {
				/* '*' is a metacharacter, since the preceding
				 * character is a either a literal or a
				 * metacharacter that does accept a '*'
				 * (. or [] or \digit (backreference)).
				 * '*' will be stored as is, because it has
				 * the same meaning in cgrep -E mode
				 * as in cgrep -G mode.
				 */
				star_ptr = q;
			}
			begin = NO;
			kill_ptr = 0;
			*q++ = *arg;
			break;

		case '\n':
			/* This is the alternation metacharacter.  Store it as is,
			 * because it has the same meaning in cgrep -E mode
			 * as in cgrep -G mode.
			 */
			starrable_ptr = star_ptr = kill_ptr = 0;
			begin = YES;
			*q++ = *arg;
			break;

		case '^':
			/* This is an beginning of line metacharacter if it
			 * is at the beginning of an expression, but is a literal
			 * otherwise.  If it is a literal, it must be escaped.
			 * If it is a metacharacter, store it as is,
			 * because it has the same meaning in cgrep -E mode
			 * as in cgrep -G mode.
			 */
			starrable_ptr = star_ptr = kill_ptr = 0;
			if (!begin) {
				/* It's a literal.  Escape it. */
				starrable_ptr = q;
				*q++ = '\\';
			}
			begin = NO;
			*q++ = *arg;
			break;

		case '$':
			/* This is an end of line metacharacter if it
			 * is at the end of an expression, but is a literal
			 * otherwise.  Since we don't yet know which, we just
			 * set "end" to D_LINE to remember the character.
			 * We store nothing at this point.
			 */
			end = D_LINE;
			begin = NO;
			starrable_ptr = star_ptr = kill_ptr = 0;
			break;

		case '#':
		case '`':
		case '(':
		case ')':
		case '+':
		case '?':
		case '|':
			/* This is a literal in cgrep -G mode.  Also,
			 * it is either a metacharacter in cgrep -E mode
			 * "()+?|" and/or a character that can be output
			 * by this routine as a metacharacter "#`()".  It
			 * must be escaped.
			 */
			begin = NO;
			starrable_ptr = q;
			star_ptr = kill_ptr = 0;
			*q++ = '\\';
			*q++ = *arg;
			break;

		case '.':
		default:
			/* This is a literal, or a metacharacter ('.')
			 * unchanged in meaning from cgrep -G mode to
			 * cgrep -E mode.  Store it as is.
			 */
			begin = NO;
			starrable_ptr = q;
			star_ptr = kill_ptr = 0;
			*q++ = *arg;
			break;

		}
	}
	/* Check for unbalanced parentheses. */
	if (paren_cnt) {
		re_error("\\( \\) imbalance");
	}
	if (q >= pat->expression + buflen) {
		re_error("-G mode conversion overflow");
	}
	/* The converted expression needs to be null terminated. */
	*q = '\0';
#ifdef	DEBUG
	printf("NEW -G EXP: %s\n", EXPRESSION);
#endif
}

/******************************************************************************

Routine:
	special()

Syntax:

	static CHAR * special(argp, q, pbegin)

	CHAR **argp;	// Pointer to pointer to first character of expression
			//   at which the special expression may be located.
	CHAR *q;	// Pointer to location at which the expansion
			//   will be stored.
	BOOL *pbegin;	// A pointer to a boolean which indicates if we are at
			//   the beginning of the expression.  Initially
			//   YES when special() is called.

Description:

	special() does the special pattern expansion for the -R or -T options.
	special() checks if the beginning of the expression pointed to by
	*argp matches the hard coded special expression.  If not,
	special() does nothing, and returns the value of input argument "q".

	If it does match, the special expansion is done, and the expansion
	is stored beginning at the location pointed to by "q".  A pointer
	to the first character after the expanded pattern is returned.
	Also, *pbegin is set to NO, to indicate to the caller that
	we are no longer at the beginning of the expression, and *argp
	is updated to point to the first character in the input expression
	after the portion that was expanded.

Global Variables:

	None

Files:
	None

Return Status:

	A pointer to the character after the expanded expression.
	If no expansion, a pointer to where the first character
	would have been stored, i.e. the value of argument 'q'.

*******************************************************************************/

static CHAR *
special(argp, q, pbegin)
CHAR **argp;
CHAR *q;
BOOL *pbegin;
{
	register CHAR *s, *e;
	register CHAR save_char;
	static first_sm = 1;

	/* If the input expression does not begin with "sm=", return.  */
	if (memcmp((char *)(*argp), "sm=", 3)) {
		return(q);
	}
	/* Point to next character in input. */
	s = *argp + 3;

	/* The remainder of the expression (terminated by a null, a new-line
	 * or a '|' metacharacter) must consist entirely of digits or
	 * of character classes containing only digits.  If not, return.
	 * Examples of valid expressions are: "sm=27", "sm=[58]", "sm=[3-5]",
	 * or "sm=2[4-7]".
	 */
	e = s + strspn((char *)s, "0123456789[-]");
	if (e == s) {
		return(q);
	}
	if (*e && *e != '|' && *e != '\n') {
		return(q);
	}

	/* OK, we will do the expansion. */
	/* No longer at beginning of expression. */
	*pbegin = NO;
	/* Advance *argp past input to be converted. */
	*argp = e;
	/* Temporarily null out the next character so that we can
	 * use string functions.
	 */
	save_char = *e;
	*e = '\0';

	/* Expand the expression.  This stuff matches most anything from
	 * the specified SM's in 5ESS ROP.  Advance "q" to point past the
	 * expansion.
	 */
 	(void)sprintf((char *)q, " [G]?[HLR]?SM[ =:]*%s[ ,:;]\n= *%s-", s, s);
	q += strlen((char *)q);
	/*  If this is the first expression expanded, add some more general
	 *  stuff that is not specific to any particular SM, but covers
	 *  SM's in general.
	 */
	if (first_sm) {
		first_sm = 0;
 		(void)sprintf((char *)q, "\n SM +[^0-9 ]");
		q += strlen((char *)q);
	}
	/* Restore the nulled out character in the input. */
	*e = save_char;
	/* Return a pointer to the next character after the output. */
	return(q);
}
