	static char sccsid[]="%Z%	output.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	output.c [%P%]

Description:

	Outputs text.

Functions:
	memchr_fwd()
	memchr_back()
	do_output()
	myexit()
	highlight()

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "re.h"
static void highlight();

/******************************************************************************

Routine:
	memchr_fwd()

Syntax:
	CHAR *memchr_fwd(p, c)

	CHAR	*p;	//  Current pointer into text
	CHAR	c;	// Character to search for

Description:

	Advances to next occurrence of specified character.
	Will stop immediately if current character matches.
	There is no check for runaway searches.

Global Variables:

	None

Files:

	None

Return Status:

	A pointer to the first character at or following "p" that matches 'c'.

*******************************************************************************/

CHAR *
memchr_fwd(p, c)
register CHAR *p;
register CHAR c;
{
	while (*p++ != c) ;
	return(--p);
}

/******************************************************************************

Routine:
	memchr_back()

Syntax:
	CHAR *memchr_back(p, c)

	CHAR	*p;	//  Current pointer into text
	CHAR	c;	// Character to search for

Description:

	Backs up to previous occurrence of specified character.
	Will stop immediately if current character matches.
	There is no check for runaway searches.

Global Variables:

	None

Files:

	None

Return Status:

	A pointer to the last character at or before "p" that matches 'c'.

*******************************************************************************/

CHAR *
memchr_back(p, c)
register CHAR *p;
register CHAR c;
{
	while (*p-- != c) ;
	return(++p);
}

/******************************************************************************

Routine:
	do_output()

Syntax:
	void do_output(begin, end, highlighting)

	CHAR *begin;		// Pointer to beginning of text to possibly
				//	be printed.  (points to NL before
				//	first character to be printed).
	CHAR *end;		// Pointer to last character to print.
	BOOL highlighting;	// Are we highlighting this output?

Description:

	Prints out specified lines (or if -c# was used, counts the specified lines).

Global Variables:

	BOOL Any_printed;	// Has any output been printed before?
	CHAR *Boffset;		// Pointer offset used for byte count (-b1).
	CHAR *Buffend;		// Pointer to end of text buffer
	CHAR *Delimiter;	// Separator string to print between
				//	non-adjacent windows.
	int Flags[];		// Command Option flags
	CHAR *Last_counted;	// Pointer to last character printed or counted
	CHAR *Last_printed;	// Pointer to last character actually printed
	BOOL L_done;		// No longer looking for matches because
				//	of -l# options
	int Line_number;	// Line number of last line printed or counted
				//	(kept accurate only if -n is being used).
	BOOL M_done;		// No longer looking for matches
	BOOL NL_search;		// Do we need to print out any auxiliary information
				//	on each line?
				//	(e.g. file name, line number, etc.)
	int Nsucc;		// Used for command exit code (errors or matches)
	int Tln;		// Line count for -c# options.

Files:

	None

Return Status:

	None:

*******************************************************************************/

void
do_output(begin, end, highlighting)
register CHAR *begin;	/* new-line before first character to "print".*/
register CHAR *end;	/* Last character to "print." */
BOOL highlighting;		/* Highlighting? */
{
	register int nc;

	/* If the last line has no new-line at the end, "end" will
	 * point to the sentinel new-line character, so we must back up.
	 * Buffend points to the last character actually in the file.
	 */
	if (end > Buffend) end = Buffend;

	/*
	 * Unless -o is in use, don't print stuff already printed.
	 */
	if (begin < Last_counted && !Flags['o']) {
		begin = Last_counted;
		if (end < Last_counted) {
			return;
		}
	}
	/* If the -n option was used, "Line_number" is the line number of the
	 * previous line printed or counted.   The end of that line is pointed
	 * to by "Last_counted".  Set "Line_number" to be the line number of
	 * the line pointed to by "begin".  We do this by counting the
	 * new-lines between "Last_counted" and "begin".
	 */
	if (Flags['n']) {
		CHAR *p = Last_counted;

		if (p < begin) {
			/* Increment "Line number". */
			while(p < begin) {
				Line_number++;
				p = memchr_fwd(p+1, NL);
			}
		} else if (p > begin) {
			while(p > begin) {
			/* Decrement "Line number".  This "backing up"
			 * occurs with the -o option when windows overlap.
			 */
				Line_number--;
				p = memchr_back(p-1, NL);
			}
		}
	}

	/* Set "Last_counted" to point to the last line we will process on this call.
	 */
	Last_counted = end;

	/* Get the number of characters to process.   Return if there are none. */
	nc  = (int)(end - begin);

	if (nc <= 0) {
		return;
	}
	/* If we are using a separator between windows, print it, unless this
	 * output begins exactly where the last output printed left off.
	 */
	if (Delimiter && begin != Last_printed) {
		(void)fprintf(stdout, "%s\n", Delimiter);
	}
	/* If we are highlighting this text, turn highlighting on. */
	if (highlighting) {
		highlight(YES);
	}

	/* Set the return code that we have matched lines. */

	Nsucc = (Nsucc == 2) ? 2 : 1;

	if (Flags['l']) {
		/* -l# was used.  We therefore need not process the output.
		 * We simply tag that we are done with the file.
		 */
		M_done = L_done = YES;
	} else if (!NL_search) {
		/* Lines can be printed directly from buffer */

		/* print the lines in one fell swoop */
		(void)fwrite((char *)(begin+1), 1, nc, stdout);
	} else {
		/* Lines need to be printed one at a time, or counted */
		register CHAR *p, *q = 0;

		/* Loop over the lines of text between "begin" and "end". */
		for (p = begin; p < end; p = q, Line_number++) {
			/* Get end of line */
			q = memchr_fwd(p+1, NL);

			/* Don't go off end of buffer. */
			if (q > Buffend) {
				q = Buffend;
			}

			/* If -c# used, just count the lines. */
			if (Flags['c']) {
				Tln++;
				continue;
			}
			/* Get number of characters in the line. */
			nc = (int)(q - p);

			/* Print filename if required. */
			if (!Flags['h']) {
				(void)fprintf(stdout, "%s:", Filename);
			}
			/* Print block number or byte offset, -b# was used.
			 * If we are printing block numbers, Flags['b'] will
			 * contain the block size.
			 */
			if (Flags['b']) {
				(void)fprintf(stdout, "%ld:",
				(unsigned long)(p - Boffset)/(unsigned int)Flags['b']);
			}
			/* Print line number if -n was used. */
			if (Flags['n']) {
				(void)fprintf(stdout, "%ld:", Line_number);
			}
			/* Print the line itself */
			(void)fwrite((char *)(p+1), 1, nc, stdout);
		}
	}
	/* If highlighting is on, turn it off. */
	if (highlighting) {
		highlight(NO);
	}
	/* Set "Last_printed" to the end of the last text actually printed,
	 * and set the "Any_printed" indicator, which indicates text has been
	 * printed.
	 */
	Last_printed = end;
	Any_printed = 1;
}

static int	high_set;

/******************************************************************************

Routine:
	myexit()

Syntax:
	void myexit(sig)

	int sig;	// If non-zero, myexit() needs to check if text
			//	highlighting mode is on, and, if so,
			//	turn it off before exiting.
			//	In this case, myexit() may have been called
			//	because of a signal, in which case,
			//	sig is the signal number.
Description:

	Exits from "cgrep" with the return value of 2, indicating an error.

	If "sig" is non-zero and highlighting has been turned on, turns off
	highlighting before exiting.  This may be necessary to restore the
	user's terminal to normal text output.

Global Variables:

	int high_set;	// If non-zero, highlighting either is currently on
			//	or has previously been on.
			// If zero, highlighting has not been turned on.

Files:

	None

Return Status:

	Does not return

*******************************************************************************/

void
myexit(sig)
int sig;
{
	if (high_set && sig) {
		highlight(NO);
	}
	exit(2);
}

/******************************************************************************

Routine:
	highlight()

Syntax:
	void highlight(sig)

	BOOL action;	// Indicates whether highlighting is to be turned on
			//	(action != 0) or off (action == 0).
Description:

	Turns highlighting on or off, as requested, but only for UN*X
	systems that support terminfo().

	Highlighting is not supported for UN*X systems that do not support
	terminfo(), or for MS_DOS.  In such a case,  highlight() is a no-op,
	although in the UN*X case it still sets or resets the global variable
	high_set.

Global Variables:

	int high_set;	// non-zero, the initialization for highlighting has
			//	already been done.

Files:

	None

Return Status:

	Does not return

*******************************************************************************/

#ifndef __MSDOS__
/* This is the UN*X version of highlight() */
#ifdef LIBNCURSES
#include <curses.h>
#include <signal.h>
#include <term.h>
#endif


static
void
highlight(action)
BOOL	action;
{

#ifdef LIBNCURSES
	/* We have UN*X terminfo(4) capability. */
	if (!high_set) {
		/* The terminfo capability has not been initialized. */
		high_set = 1;
		/* Catch hangup and interrupt signals, so highlighting won't
		 * be left on inadvertently.
		 */
		(void)signal(SIGHUP, myexit);
		(void)signal(SIGINT, myexit);

		/* This does the initialization required to use terminfo(4).
		 * See curs_terminfo(3X) for a description of setupterm().
		 * The second argument of setupterm is the file descriptor
		 * (1 for standard output).
		 * NOTE: resetterm() does not seem to be documented currently,
		 * and may be obsolete.  However the call has been left in
		 * because "If it's not broken, don't fix it."
		 */
		setupterm(0, 1, 0);
		/* But, it's not in ncurses! */		
		/* resetterm(); */
		reset_shell_mode();
	}

	/* This turns highlighting on or off by use of the terminfo
	 * capability.  See curs_terminfo(3X) for a description of putp().
	 */
	if (action) {
		putp(enter_standout_mode);
	} else {
		putp(exit_standout_mode);
	}
	/* Output a carriage return.  This is necessary to reset the tab
	 * stops to their proper location, in case the next line of text
	 * contains a tab.  Testing has shown that the output of non-printing
	 * escape sequences by putp() can otherwise affect the tab stops.
	 */
	(void)fprintf(stdout,"%c",'\r');
#else
	/* No terminfo capability.  Just turn high_set on and off,
	 * in case we should ever need it.
	 */
	high_set = (int)action;
#endif
}
#else

/* This is the MS_DOS dummy version of highlight(). */

void
highlight()
{
}
#endif
