	static char sccsid[]="%Z%	reexec.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	reexec.c [%P%]

Description:

	Contains the routines for matching regular expressions.

Functions:

	re_reinit();
	re_reexec();
	re_rbexec();
	get_beg();
	get_end();
	nextstate();
	stateof();
	followstate();
	match();
	efollow();
	follow();
	first();
	addstate();

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include	"re.h"

CHAR	bittab[] = { 128, 64, 32, 16, 8, 4, 2, 1 };

/* Define NO_INSERT to be the largest unsigned short (i.e., all 1's). */

#define NO_INSERT (unsigned short)~0

#ifdef DEBUG
static void eg_spr_f();
static void eg_spr_c();
void eg_epr();
int debug = 3;
#endif

/* PSFID_PTR(rps, n) extracts a pointer to PSFID <n> from the POSITION SET
 * pointed to by <rps>. */

/******************************************************************************

MACRO:
	PSFID_PTR()

Syntax:

	PSFID *PSFID_PTR(_ps, _id)

	POSITION_SET *_ps;	// Pointer to POSITION_SET
	ID _id;			// ID of PSFID


Description:

	PSFID(), a MACRO, returns a pointer to the PSFID that has ID <_id>
	in POSITION_SET <_ps>.

Return Status:

	Returns pointer to PSFID desired.

*******************************************************************************/

#define	PSFID_PTR(_ps, _id) (&((_ps)->base[_id]))

/******************************************************************************

MACRO:
	CLEAR()

Syntax:

	void CLEAR(_r, _ps)

	RE_RE	*_r;	// Pointer to compiled regular expression structure.
	POSITION_SET *rps;	// Pointer to POSITION_SET to be cleared.

Description:

	CLEAR(), a MACRO, initializes the POSITION_SET pointed to by <_ps>
	to an empty POSITION_SET.

Return Status:

	None

*******************************************************************************/

#define CLEAR(_r, _ps)	(void)memset((char *)(_ps)->base, 0, (int)((_r)->maxid*sizeof(PSFID))), (_ps)->count = 0, (_ps)->last = -1

/******************************************************************************

MACRO:
	SET()

Syntax:

	void SET(_ps, _id, _cost, _ins_c0st)

	POSITION_SET *_ps;  // Pointer to POSITION_SET
	ID _id;		    // ID of PSFID to be entered into POSITION_SET
	short _cost;	    // Value of <cost> for new POSITION_SET entry.
	short _ins_cost;    // Value of <ins_cost> for new POSITION_SET entry.


Description:

	SET(), a MACRO, enters a PSFID into a POSITION_SET, or updates 
	an existing PSFID in a POSITION_SET.

	CASE 1: If the PSFID with ID <_id> is not already in POSITION_SET <_ps>,
	then add it to POSITION_SET <_ps> with the values specified.

	CASE 2: The PSFID with ID <_id> is already in the POSITION_SET.
	The accumulated cost, <cost>, will be set to the minimum of the
	specified value and the previous value.
	The insert cost, <ins_cost>, will be set so that the sum of the
	new values of <ins_cost> and <cost> will be equal to the minimum of
	a) the sum of the previous values of <ins_cost> and <cost>
	or
	b) the sum of the specified values <_ins_cost> and <_cost>.
	This is done in order to obtain the minimum cost after the next
	error if the next error is an insertion error.

Return Status:

	None

*******************************************************************************/

#define	SET(_ps, _id, _cost, _ins_cost) { \
	register PSFID *_q = PSFID_PTR(_ps, _id);\
	if (_q->previd == 0) {\
		(_ps)->count++;\
		_q->cost = (_cost);\
		_q->ins_cost = (_ins_cost);\
		_q->previd = (_ps)->last;\
		(_ps)->last = (_id);\
	} else {\
		unsigned _c1 = _q->cost + (unsigned)_q->ins_cost;\
		unsigned _c2 = (_cost) + (unsigned)(_ins_cost);\
		unsigned _cmin = MIN(_c1, _c2);\
		_q->cost = MIN(_q->cost, (_cost));\
		_q->ins_cost = _cmin - _q->cost;\
	}}

/* Empty_state is a state of the Finite State Machine, indicating no matches are
 * possible.  Its contents are never accessed.  However, a STATE pointer can
 * point to it.  This must be checked for in any anchored match, in which case
 * this becomes a possible state.
 */
STATE Empty_state;

static STATE *stateof(), *addstate();
static BOOL match(), first();
static void follow(), followstate();
static CHAR *get_end();


/******************************************************************************

Routine:
	re_reinit()

Syntax:

	void re_reinit(r)

	RE_RE	*r;	// Pointer to compiled regular expression structure.

Description:

	If regular expression matching is to be used for a pattern,
	"re_reinit()" must be called after the regular expression is
	compiled (by "re_recomp()"), but before any text matching
	is attempted (by "re_reexec()" or "re_rbexec()").  "re_reinit()"
	performs necessary PRE_MATCHING initialization of a compiled
	regular expression (RE_RE) structure before it can be used for
	text matching.

	If backreferencing is in use, "re_brinit()" is called instead
	of re_reinit() for PRE_MATCHING initialization.  However,
	"re_brinit()" indirectly calls "re_reinit()" to aid in the
	initialization of each subexpression.


Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

void
re_reinit(r)
register RE_RE *r;
{
	unsigned int n;

	/* Allocate memory for two arrays of PSFID's, used in POSITION_SETS. */
	n = r->maxid * sizeof(PSFID);
	r->ps.base = (PSFID *)cgmalloc(n, "ps base");
	r->fpos.base = (PSFID *)cgmalloc(n, "psf base");

	/* Use the "ps" POSITION_SET temporarily to calculate the initial STATE.
	 * Initialize the POSITION_SET to empty.
	 */
	CLEAR(r, &r->ps);

	/* Add the appropriate "lit" expressions to the POSITION_SET
	 * for the initial state.
	 */
	if (!first(r, &r->ps, r->root->l, (short)0, NO_INSERT)) {
		/*
		 * The pattern matches the null string!!!!
		 * Add EOP to the POSITION_SET for the initial state.
		 */
		SET(&r->ps, r->root->id, (short)0, NO_INSERT);
		/*PR("first of root->l == 0, b=%s\n", b);*/
	}
	/* Allocate memory for the general array of POSITION_SETs
	 * (r->posbase) and for the general array of STATEs (r->states).
	 * Also initialize some related entries in "r".
	 */
	eg_stateinit(r);

	/* Create the initial state.  The PSFID's from POSITION_SET "r->ps"
	 * will be stored in "r->posbase" as FID's.
	 */
	(void)addstate(r, &r->ps);

	/* Prevent the FID's for the initial state from being
	 * flushed from "r->posbase" if states have to be flushed.
	 */
	eg_savestate(r);
	return;
}

/******************************************************************************

Routine:
	re_reexec()

Syntax:

	BOOL re_reexec(pat, fcn, b, e)

	PATTERN *pat;	// Pointer to pattern to be matched
	BOOL (*fcn)();	// Pointer to subsequent pattern matching function
			//	(0 if this is final algorithm used)
	CHAR *b;	// Pointer to first character of text to be searched
	CHAR *e;	// Pointer to last character of text to be searched

Description:

	re_reexec() is the text matching routine for regular expressions,
	when the text is being searched in the forward direction.

	re_reexec() calls grepmatch() for each match that is found, to store
	any match information that needs to be stored.  After a match is found,
	re_reexec() will continue to search for matches, unless grepmatch()
	indicates by its return value that the search should stop.

	re_reexec() does not handle backreferencing, for which re_brexec()
	is used instead.  It does not handle anchored matches, which are
	currently only used within backreferencing, for which re_brexec()
	is used, or for finding the length of a match, for which
	get_beg() or get_end() is used.

Global Variables:

	None

Files:
	None

Return Status:

	YES 	// At least one match was found, and the search was then
		//   terminated because grepmatch() indicated that
		//   it should be terminated.

	NO	// The entire text was searched for matches.   Any number
		//   of matches may have been found.

*******************************************************************************/

BOOL
re_reexec(pat, fcn, b, e)
PATTERN *pat;
BOOL (*fcn)();
register CHAR *b, *e;
{
	register RE_RE *r = pat->re_ptr;
	register STATE *s, *t;
	CHAR *nb, *ne;
	/* Save the start of the text to search */
	CHAR *save_b = b;

#ifdef	DEBUG
	if (TRACE(10)) {
		CHAR buf[EPRINTSIZE];
		eg_epr(r->root, buf);
		PR("expr='%s'\n", buf);
		PR("attempting match at input '%s'\n", b);
	}
#endif
again:
	/* Start at the initial state. */
	s = &r->states[0];

	/* There is no need to check for a match before we read
	 * any characters.  If the initial state is a match, the
	 * pattern matches the null string, every state matches!
	 */

	/* Loop over text characters */
	for (; b <= e; b++) {
		register CHAR c = r->cmap[*b];
#ifdef	DEBUG
		if (TRACE(4))
			PR("state %d: char '%c'\n", s-r->states, *b);
#endif
		/* A character has been scanned and mapped.  Advance
		 * to the next state of the finite-state machine.
		 */
		if ((t = s->tab[c]) != 0) {
			/* This transition has occurred before. */
			s = t;
		} else {
			/* This transition has not occurred before.
			 * The new state must be found or calculated.
			 */
			s = nextstate(r, s, (int)c);
		}
		/* Do we now have a match? */
		if (s->is_match) {
			CHAR *beg = b;
			/* We have a match.  Get the beginning of the match,
			 * if we care.
			 */
			if (r->length) {
				/* Matches are fixed length */
				beg = b + 1 - r->length;
			} else if (pat->count_ptr) {
				/* Matches are not fixed length,
				 * and we need the beginning.
				 */
				beg = get_beg(pat->count_ptr, save_b, b);
			} else if (r->min_length) {
				/* We don't really care where the beginning is.
				 * Just assume a two character long match.
				 * In this way, if the last character matched
				 * is a new-line, we will get the previous
				 * character, while if the last character matched
				 * is not a new-line, we will never get a
				 * character on the previous line.
				 */
				beg = b - 1;
			}
#ifdef	DEBUG
			if (TRACE(6)) {
				PR("match ending at input '%s' count '%d'\n", b+1, b+1-beg);
			}
#endif
			/* grepmatch() will do whatever is necessary to store
			 * the match, if indicated.
			 */
			if (grepmatch(pat, fcn, save_b, e, beg, b, &nb, &ne)) {
				/* A return value of YES from getversion
				 * indicates we should stop looking for matches.
				 * This happens only if we are counting
				 * matches and have reached the maximum,
				 * or if we are doing line-by-line matching
				 * because of -v.
				 */
				return(YES);
			}
			/* Keep looking for more matches. */
			if (ne > b + 1) {
				/* The match had been extended by grepmatch()
				 * to the end of the current line of text.
				 * There is no need to check for more
				 * matches on the same line.
				 */
				b = ne;
				/* We start again with the initial state,
				 * since we started a new line.
				 */
				goto again;
			}
		}
	}
#ifdef	DEBUG
	if (TRACE(3)) {
		CHAR buf[EPRINTSIZE];

		eg_epr(r->root, buf);
		PR("pat = %s\n", buf);
	}
#endif
	/* We have searched the entire text. */
	return(NO);
}

/******************************************************************************

Routine:
	re_rbexec()

Syntax:

	BOOL re_rbexec(pat, fcn, b, e)

	PATTERN *pat;	// Pointer to pattern to be matched
	BOOL (*fcn)();	// Pointer to subsequent pattern matching function
			//	(0 if this is the final algorithm used)
	CHAR *b;	// Pointer to first character of text to be searched
	CHAR *e;	// Pointer to last character of text to be searched

Description:

	re_rbexec() is the text matching routine for regular expressions,
	when the text is being searched in the BACKWARD direction.

	re_rbexec() calls grepmatch() for each match that is found, to store
	any match information that needs to be stored.  After a match is found,
	re_rbexec() will continue to search for matches, unless grepmatch()
	indicates by its return value that the search should stop.

	re_rbexec() does not handle backreferencing, for which re_brexec()
	is used instead.  It does not handle anchored matches, which are
	currently only used within backreferencing, for which re_brexec()
	is used, or for finding the length of a match, for which
	get_beg() or get_end() is used.

Global Variables:

	None

Files:
	None

Return Status:

	YES 	// At least one match was found, and the search was then
		//   terminated because grepmatch() indicated that
		//   it should be terminated.

	NO	// The entire text was searched for matches.   Any number
		//   of matches may have been found.

*******************************************************************************/

BOOL
re_rbexec(pat, fcn, b, e)
PATTERN *pat;
BOOL (*fcn)();
register CHAR *b, *e;
{
	register RE_RE *r = pat->re_ptr;
	register STATE *s, *t;
	CHAR *nb, *ne;
	/* Save the end of the text to search */
	CHAR *save_e = e;

#ifdef	DEBUG
	if (TRACE(10)) {
		CHAR buf[EPRINTSIZE];
		eg_epr(r->root, buf);
		PR("expr='%s'\n", buf);
		PR("attempting match at input '%s'\n", b);
	}
#endif
again:
	/* Start at the initial state. */
	s = &r->states[0];

	/* There is no need to check for a match before we read
	 * any characters.  If the initial state is a match, the
	 * pattern matches the null string, every state matches!
	 */

	/* Loop over text characters */
	for (; e >= b; e--) {
		register CHAR c = r->cmap[*e];
#ifdef	DEBUG
		if (TRACE(4))
			PR("state %d: char '%c'\n", s-r->states, *b);
#endif
		/* A character has been scanned and mapped.  Advance
		 * to the next state of the finite-state machine.
		 */
		if ((t = s->tab[c]) != 0) {
			/* This transition has occurred before. */
			s = t;
		} else {
			/* This transition has not occurred before.
			 * The new state must be found or calculated.
			 */
			s = nextstate(r, s, (int)c);
		}
		/* Do we now have a match? */
		if (s->is_match) {
			CHAR *end = e;
			/* We have a match.  Get the end of the match,
			 * if we care.
			 */
			if (r->length) {
				/* Matches are fixed length */
				end = e + r->length - 1;
			} else if (pat->count_ptr) {
				/* Matches are not fixed length,
				 * and we need the end.
				 */
				end = get_end(pat->count_ptr, e, save_e);
			} else if (r->min_length) {
				/* We don't really care where the beginning is.
				 * Just assume a two character long match.
				 * In this way, if the first character matched
				 * is a new-line, we will get the next
				 * character, while if the last character matched
				 * is not a new-line, we will never get a
				 * character on the next line.
				 */
				end = e + 1;
			}
#ifdef	DEBUG
			if (TRACE(6)) {
				PR("match beginning at input '%s' count '%d'\n", e-1, end+1-e);
			}
#endif
			/* grepmatch() will do whatever is necessary to store
			 * the match, if indicated.
			 */
			if (grepmatch(pat, fcn, b, save_e, e, end, &nb, &ne)) {
				/* A return value of YES from getversion
				 * indicates we should stop looking for matches.
				 * This happens only if we are counting
				 * matches and have reached the maximum,
				 * or if we are doing line-by-line matching
				 * because of -v.
				 */
				return(YES);
			}
			/* Keep looking for more matches. */
			if (nb < e - 1) {
				/* The match had been extended by grepmatch()
				 * to the beginning of the current line of text.
				 * There is no need to check for more
				 * matches on the same line.
				 */
				e = nb;
				/* We start again with the initial state,
				 * since we started a new line.
				 */
				goto again;
			}
		}

	}
#ifdef	DEBUG
	if (TRACE(3)) {
		CHAR buf[EPRINTSIZE];

		eg_epr(r->root, buf);
		PR("pat = %s\n", buf);
	}
#endif
	/* We have searched the entire text. */
	return(NO);
}

/******************************************************************************

Routine:
	get_beg()

Syntax:

	CHAR *get_beg(r, b, e)

	RE_RE *r;	// Pointer to regular expression that is a reversed
			//   copy of the one that was matched
	CHAR *b;	// Pointer to first character of text that was searched
	CHAR *e;	// Pointer to last text character of the match

Description:

	get_beg() determines the longest length of text that matches a
	regular expression, given that, while matching in the forward
	direction, a match has been found that terminates at "e".

	get_beg() begins at "e", and searches in the reverse direction
	(i.e. backward) for the longest match to a regular expression "r",
	which is a reversed (i.e. backward) copy of the original forward
	regular expression.

	The search for this match is "anchored", that is, only matches
	that start at "e" itself are found.  This is implemented by the fact
	that "r->anchored" has previously been set to YES in set_count_ptr().

Global Variables:

	Empty_state	// A state of the Finite State Machine that indicates
			//   we no longer can match the regular expression.
			//   This state occurs only for anchored matches.
			//   For example, if we are trying to match the
			//   regular expression "ab*c", and the first character
			//   is not an "a", we arrive at Empty_state, because
			//   anchoring requires that the match begin at the
			//   first character.

Files:
	None

Return Status:

	Returns a pointer to the beginning of the longest possible match
	to "r" that ends at "e".

*******************************************************************************/

CHAR *
get_beg(r, b, e)
register RE_RE *r;
register CHAR *b, *e;
{
	register STATE *s, *t;
	/* Initialize the beginning of the match to one character AFTER the end.
	 * This is how a zero-length match is represented.
	 */
	register CHAR *beg = e+1;
	register CHAR c;

	/* Start at the initial state. */
	s = &r->states[0];

	/* Loop backwards through the text.  Don't back up beyond "b", the
	 * first character of the text that was searched for the original match.
	 * There is no need to check that the initial state is a match,
	 * because we have already initialized "beg" for a zero length match.
	 */
	for (; e >= b; e--) {
		/* Get one text character, and map it. */
		c = r->cmap[*e];

		/* A character has been scanned and mapped.  Advance
		 * to the next state of the finite-state machine.
		 */
		if ((t = s->tab[c]) != 0) {
			/* This transition has occurred before. */
			s = t;
		} else {
			/* This transition has not occurred before.
			 * The new state must be found or calculated.
			 */
			s = nextstate(r, s, (int)c);
		}

		/* If we reach Empty_state, stop looking for matches.  */
		if (s == &Empty_state) break;

		/* If we have a match, set "beg" to point to the first
		 * character.  We still keep looping, because there
		 * may still be a longer match.
		 */
		if (s->is_match) beg = e;
	}
	return(beg);
}

/******************************************************************************

Routine:
	get_end()

Syntax:

	CHAR *get_end(r, b, e)

	RE_RE *r;	// Pointer to regular expression that is a reversed
			//   copy of the backward one that was matched
	CHAR *b;	// Pointer to first text character of the match
	CHAR *e;	// Pointer to last character of text that was searched

Description:

	get_end() determines the longest length of text that matches a
	regular expression, given that, while matching in the BACKWARD
	direction, a match has been found that terminates at "b".

	get_end() begins at "b", and searches in the reverse direction
	(i.e. forward) for the longest match to a regular expression "r",
	which is a reversed (i.e. forward) copy of the original backward
	regular expression.

	The search for this match is "anchored", that is, only matches
	that start at "b" itself are found.  This is implemented by the fact
	that "r->anchored" has previously been set to YES in set_count_ptr().

Global Variables:

	Empty_state	// A state of the Finite State Machine that indicates
			//   we no longer can match the regular expression.
			//   This state occurs only for anchored matches.
			//   For example, if we are trying to match the
			//   regular expression "ab*c", and the first character
			//   is not an "a", we arrive at Empty_state, because
			//   anchoring requires that the match begin at the
			//   first character.

Files:
	None

Return Status:

	Returns a pointer to the beginning of the longest possible match
	to "r" that ends at "e".

*******************************************************************************/

static CHAR *
get_end(r, b, e)
register RE_RE *r;
register CHAR *b, *e;
{
	register STATE *s, *t;
	/* Initialize the end of the match to one character BEFORE the beginning.
	 * This is how a zero-length match is represented.
	 */
	register CHAR *end = b-1;
	register CHAR c;

	/* Start at the initial state. */
	s = &r->states[0];

	/* Loop forward through the text.  Don't advance beyond "e", the
	 * last character of the text that was searched for the original match.
	 * There is no need to check that the initial state is a match,
	 * because we have already initialized "end" for a zero length match.
	 */
	for (; b <= e; b++) {
		/* Get one text character, and map it. */
		c = r->cmap[*b];

		/* A character has been scanned and mapped.  Advance
		 * to the next state of the finite-state machine.
		 */
		if ((t = s->tab[c]) != 0) {
			/* This transition has occurred before. */
			s = t;
		} else {
			/* This transition has not occurred before.
			 * The new state must be found or calculated.
			 */
			s = nextstate(r, s, (int)c);
		}

		/* If we reach Empty_state, stop looking for matches.  */
		if (s == &Empty_state) break;

		/* If we have a match, set "end" to point to the last
		 * character.  We still keep looping, because there
		 * may still be a longer match.
		 */
		if (s->is_match) end = b;
	}
	return(end);
}

/******************************************************************************

Routine:
	nextstate()

Syntax:

	STATE *nextstate(r, s, a)

	RE_RE *r;	// Pointer to the regular expression we are matching
	STATE *s;	// Current STATE
	int a;		// Mapped text character being processed

Description:

	nextstate() determines the next STATE based on the current state,
	the text character being processed, and whether or not the match
	is "anchored" to begin at a specific character.   The new STATE
	is created if it is not already stored.  The transition table
	entry from the current state to the new state is filled in,
	and a pointer tot he new state is returned.

	nextstate() does not check to see if the transition to the new
	state is already stored in the transition table.  This should
	be done by the caller.  If the translation table entry already
	exists, there is no need to call nextstate().  We do this in each
	caller to save the function call overhead, which can at least
	quadruple the execution time in some cases.

Global Variables:

	Empty_state	// A state of the Finite State Machine that indicates
			//   we no longer can match the regular expression.
			//   This state occurs only for anchored matches.
			//   For example, if we are trying to match the
			//   regular expression "ab*c"with an anchored match,
			//   and the first character is not an "a", we arrive
			//   at Empty_state, because  anchoring requires that
			//   the match begin at the first character.

Files:
	None

Return Status:

	Returns a pointer to the next state of the finite state machine.

*******************************************************************************/

STATE *
nextstate(r, s, a)
register RE_RE *r;
register STATE *s;	/* Current STATE */
int a;			/* Current text character */
{
	register STATE *news;
	BOOL flushed = NO;
	STATE *old_states;
	
	/* Clear the temporary position set. */
	CLEAR(r, &r->ps);

	/* Store in the position set any "lit" expression that can
	 * follow a "lit" expression matching the current text character (a)
	 * in a possible match to the entire pattern.  For example, if the
	 * regular expression to be matched is "ghij|phqrs|hxy", the
	 * current text character (a) is 'h', and the previous character
	 * was 'g', the 'i' and the 'x' would be stored in the position set,
	 * if the match is unanchored.
	 * The 'q' is not stored, because even though it follows an 'h',
	 * THAT 'h' is not relevant because it follows a mismatched
	 * character, and therefore cannot be part of a match at the current
	 * place in the text.
	 * If the match is anchored to start at the 'g' in the text,
	 * only the 'i' would be stored in the position set.
	 */
	followstate(r, s, a, &r->ps);

#ifdef	DEBUG
	if (TRACE(5)) {
		CHAR buf[2000];
		ppr(&r->ps, buf);
		PR("nextstate(%d, '%c'): found %s\n", s-r->states, a, buf);
	}
#endif
	/* See if new the POSITION_SET is empty. */
	if (r->ps.count == 0) {
		/* The new position set is empty.
		 * If the match is anchored, we can no longer have a match.
		 * Set the new state to Empty_state.
		 * If the match is unanchored, we can start over again
		 * looking for new matches.
		 * Set the new state to the initial state.
		 */
		news = (r->anchored ? &Empty_state : &r->states[0]);
	} else if ((news = stateof(r, &r->ps)) == 0) {
		/* The new POSITION_SET is not empty.  Furthermore,
		 * the "stateof()" call found that the new STATE is
		 * not already in memory.  We must add the new state
		 * to the buffer.
		 *
		 * WARNING:  If the number of states gets very large,
		 * the buffer containing the states will be flushed
		 * before creating the new state.  We need to check
		 * for this explicitly.
		 */
		/* Save the number of times the state buffer has been flushed. */
		int prev_flushes = r->flushed; 
		/* Create the new state and add it to the buffer. */

		old_states = r->states;
		news = addstate(r, &r->ps);

		/* PSF 06/24/04 - It turns out that addstate() 
		 * can have a side effect of moving r->states if
		 * it needs to grow.   However, since s points
		 * to an entry within r->states, it's left pointing
		 * to the old location, unless we patch it up here.
		 * An architectural change might be a better way to
		 * handle this, but this way seems safer as it 
		 * doesn't change the way the code executes in the 
		 * normal case.
		 */
		if ( r->states != old_states) {
			int index = s - old_states;
			s = &(r->states[index]);
		}
		
		/* Check if state "s" was flushed from the buffer.  If the
		 * buffer was flushed, all states except the initial
		 * state were flushed.
		 */
		if (r->flushed > prev_flushes && s != &r->states[0]) {
			flushed = YES;
		}
	}
	/* Set the transition table entry for the transition from the old
	 * state "s" to the new state "news".  However, we do this only
	 * if "s" has not been flushed from the state buffer.
	 */
	if (flushed == NO) {
		s->tab[a] = news;
	}
#ifdef	DEBUG
	if (TRACE(5)) {
		PR("nextstate(%d, '%c'): returning %ld\n", s-r->states, a, news);
	}
#endif
	/* Return a pointer to the new state, */
	return(news);
}

/******************************************************************************

Routine:
	stateof()

Syntax:

	static STATE *stateof(r, ps)

	RE_RE *r;		// Pointer to the regular expression we are matching
	POSITION_SET *ps;	// Pointer to POSITION_SET of next state.

Description:

	stateof() is called by nextstate() to determine if the next state
	is identical to a state already stored in the buffer.  If so, a
	pointer to that state is returned.

Global Variables:

	None

Files:
	None

Return Status:

	Returns a pointer to the state, if the state is in memory.
	Returns 0 if the state is not in memory.

*******************************************************************************/

static STATE *
stateof(r, ps)
RE_RE *r;
register POSITION_SET *ps;
{
	register STATE *s;
	register int i;
	register FID *p, *e;

	/* Loop over all the states in memory. */
	for (i = 0, s = r->states; i < r->nstates; i++, s++) {
		/* Check if the FID count of this state matches that
		 * of our POSITION_SET.
		 */
		if (s->npos == ps->count) {
			/* The counts match.  Therefore, if every FID in the
			 * state is also in the POSITION_SET, they are all
			 * identical.
			 * Loop over the FID's in the state.
			 */
			for (p = s->pos+r->posbase, e = p+s->npos; p < e; p++) {
				/* We have a FID from the state.  Get the
				 * entry for the corresponding "lit"
				 * expression from the POSITION_SET.
				 */
				register PSFID *q = &ps->base[p->id];

				/* In order for the state and POSITION_SET
				 * to be identical, the "lit" must be in
				 * the POSITION_SET (q->previd != 0).  Also,
				 * if approximate matching is in use,
				 * the accumulated cost of mismatches must
				 * be the same (q->cost == p->cost),
				 * and the cost of inserting a mismatched
				 * character ahead of the next matched character
				 * must be the same (q->ins_cost ==  p->ins_cost).
				 */
				if (q->previd != 0 && q->cost == p->cost &&
				    q->ins_cost == p->ins_cost) {
					/* This FID matches the POSITION_SET.
					 * Now, check the next FID in the state.
					 */
					continue;
				}
				/* Either the "lit" from the state is not
				 * in the POSITION_SET, or the error costs
				 * for approximate matching are not the same.
				 * This state does not match the POSITION_SET.
				 * Try the next state.
				 */
				goto next;
			}
			/* All the FID's match.  This state exactly corresponds
			 * to the position set.  Return a pointer to the state.
			 */
			return(s);
		}
	next:;
	}
	/* No state in memory matched the POSITION_SET. */
	return((STATE *)0);
}


/******************************************************************************

Routine:
	followstate()

Syntax:
	static void followstate(r, s, a, ps)

	RE_RE *r;	// Pointer to the regular expression we are matching
	STATE *s;	// Current STATE
	int a;		// Mapped text character being processed
	POSITION_SET *ps;
			// Pointer to struct in which to store the
			//   POSITION_SET of the next state.

Description:

	Given a regular expression (r), the current state (s) of the finite
	state machine, and the next character to be processed (a), followstate()
	determines the POSITION_SET of the next state, and stores it
	in "ps".

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
followstate(r, s, a, ps)
register RE_RE *r;
register STATE *s;
int a;
POSITION_SET *ps;
{
	register int j;
	register FFID *q, *eq;
	EXPR *e;
	FOLLOW_SET *f;
	short cost;
	unsigned short ins_cost;

#ifdef DEBUG
	printf("TEXT_CHAR %c\n", a);
#endif
	/* Loop over the FID's in the POSITION_SET for the current state. */
	for (j = s->pos; j < (s->pos + s->npos); j++) {

		/* Get the "lit" expression for this FID. */
		e = r->ptr[r->posbase[j].id];

		/* Get the cost of text mismatches accumulated so far.
		 * Needed for approximate matching.  Will be 0 if not
		 * doing approximate matching.
		 */
		cost = r->posbase[j].cost;
		/* Get the cost of an an extraneous character in the text
		 * before "a".  Needed for approximate matching.  Will be
		 * NO_INSERT if such an insertion is not allowed or if not doing
		 * approximate matching.
		 */
		ins_cost = r->posbase[j].ins_cost;
#ifdef DEBUG
		printf("STATE_CHAR %c, COST: %d\n", e->lit, cost);
#endif
		/* This next block of code gets executed if we are allowed
		 * to treat text character "a" as an insertion error
		 * before the next matched character (approximate matching).
		 * If not, "ins_cost" will be NO_INSERT.
		 */
		if (ins_cost != NO_INSERT) {
			/* It may be allowed to treat "a" as an insertion error.
			 * However, if we do, the accumulated mismatch cost
			 * must be incremented by "ins_cost".
			 * If "ins_cost" is 0, a new-line still counts as one.
			 */
			short new_cost = cost + ins_cost;
			if (!ins_cost && a == NL) new_cost++;

			/* If we haven't exceeded the cost limit, yet, we
			 * allow "a" to be an insertion error.
			 */
			if (new_cost <= r->cost_limit) {
				/* Since "a" can be treated as an insertion
				 * error, the FID in the current state "s"
				 * will also be in the POSITION_SET for the
				 * new set, albeit with the new possibly
				 * higher mismatch cost.
				 */
				SET(ps, e->id, new_cost, ins_cost);
#ifdef DEBUG
				printf("CHAR %c PREV_COST %d NEW_COST %d\n",
					a, cost, new_cost);
#endif
			}
		}
		/* Now check if character "a" does NOT match "lit" expression "e". */
		if (!match(e, a)) {
			/* No match.   Check if we are allowed a substitution
			 * mismatch.
			 */
			if (!e->n_bracket) {
				/* We are approximate matching, and "lit" "e"
				 * is not inside angle brackets.  We can
				 * treat the mismatch as a match, provided
				 * we increment the accumulated mismatch cost
				 * by the cost of a substitution error, and
				 * provided the cost does not exceed the limit.
				 */
				cost += r->cost_sub;
				if (cost > r->cost_limit) {
					/* We are not allowed this substitution,
					 * because our cost has exceeded the limit.
					 * We don't match "lit" "e".
					 */
					continue;
				}
			} else {
				/* We are not allowed a substitution.
				 * We don't match "lit" "e".
				 */
				continue;
			}
		}
#ifdef DEBUG
		printf("STATE_CHAR %c, NEW_COST: %d\n", e->lit, cost);
#endif
		/* We have matched "lit" "e". Get a pointer to the set of FID's
		 * for "lit" expressions that may follow (precede) "e" in a match
		 * if we are matching text in the forward (backward) direction.
		 */
		f = (r->backward ? &e->b : &e->f);
		/* This set is not filled in until the first time it is needed.
		 * If the set has not yet been filled in, efollow() will fill it in.
		 */
		if (f->follow == 0) {
			efollow(r, e, f);
		}
		/* Loop through the set of FIDs for "lit" expressions that may
		 * follow (or precede, if we are matching backwards) "e" in a match.
		 */
		for (q = f->follow, eq = q + f->flen; q < eq; q++) {
			/* q->cost contains the incremental cost due to
			 * mismatches that we must pay to have this FID
			 * follow "e".  Check that this additional cost does
			 * not drive the total cost over the limit.
			 */
			short new_cost = cost + q->cost;
			if (new_cost <= r->cost_limit) {
				/* Add this FID to the position set for the
				 * next state, storing the new cost of
				 * accumulated mismatches.
				 */
				SET(ps, q->id, new_cost, q->ins_cost);
#ifdef DEBUG
				printf("CHAR %c FC %c PREV_COST %d FOLLOW_COST %d SET_TOT %d\n",
					a, r->ptr[q->id]->lit,
					cost, q->cost, new_cost);
#endif
			}
		}
	} /* End of loop over the FID's in the POSITION_SET for the current state. */

	/* We may not be finished.  If state "s" is not the initial state,
	 * "a" is not the first character of the potential match.  We have
	 * added to the position set only those FID's that would
	 * extend a preexisting potential match.  However, if the match is not
	 * anchored, "a" may also be the first character of a new match.
	 * So, if "s" is not the initial state, and the match is not anchored,
	 * we call "followstate" again, passing in the initial state, rather
	 * than the current state.  This will add to the position set for the
	 * new state those FID's that can follow "a" at the beginning of a new
	 * match.
	 *
	 * Here is an example (assume no approximate matching):
	 * The regular expression we are matching is "fghi+j|pqrst".
	 * The position set of the initial state consisted of 'f' and 'p'.
	 * The first text character was an 'f'.  The position set of the
	 * subsequent state consisted only of 'g'.  Now, the second text
	 * character read was a 'p'.  Now, when followstate is called,
	 * the position set calculated on this first pass will be empty,
	 * because 'p' does not match 'g'.  In order to get 'q' into the
	 * position set, we must check if 'p' might be the first character of
	 * a new match.  This is the purpose of this additional call to
	 * followstate().
	 */
	if (s != &r->states[0] && !r->anchored) {
		followstate(r, &r->states[0], a, ps);
	}
}

/******************************************************************************

Routine:
	match()

Syntax:

	static BOOL match(e, a)

	EXPR *e;	// Pointer to a "lit" expression to be
			//   compared with a text character.

	int a;		// The text character.

Description:

Global Variables:

	None

Files:
	None

Return Status:

	YES	// A match
	NO	// No match

*******************************************************************************/

static BOOL
match(e, a)
register EXPR *e;
int a;
{
	switch(e->type) {

	/* A Dot expression matches any character except a new-line. */
	case Dot:	return((BOOL)(a != NL));

	/* A Literal expression matches only one character. */
	case Literal:	return((BOOL)(a == e->lit));

	/* A Charclass expression matches a set of characters.  The
	 * character masks are stored in e->r.
	 */
	case Charclass:	return((BOOL)(CCL_CHK((CHAR *)e->r, a)));
	}
	/* Return NO for any other expression type. */
	return(NO);
}

/******************************************************************************

Routine:
	efollow()

Syntax:

	void efollow(r, e, f)

	RE_RE *r;	// Pointer to the regular expression we are matching
	EXPR *e;	// Current "lit" expression
	FOLLOW_SET *f;	// Pointer to FOLLOW_SET in which to store the
			//   set of FID's that can possibly follow "e"
			//   in a match to text.
			// If NULL, no FOLLOWSET is created.

Description:

	efollow() calls  follow() to store all the FID's for "lit" expressions
	that can follow <e> in a match into a POSITION_SET.  Then, if <f>
	is not null, efollow() allocates memory for storing the FID's in the
	FOLLOW_SET pointed to by <f>, and then copies the information
	from the POSITION_SET into the FOLLOW_SET.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

void
efollow(r, e, f)
register RE_RE *r;
register EXPR *e;
FOLLOW_SET *f;
{
	register ID id;
	register FFID *p;
	register PSFID *q = 0;
	register POSITION_SET *fpos = &r->fpos;

	/* Clear the "fpos" POSITION_SET. */
	CLEAR(r, fpos);

	/* Call follow() to fill in the position set with the
	 * possible following FID's.
	 */
	follow(r, fpos, e, (short)0, e->n_bracket);

	/* Now, copy the information from the POSITION_SET <fpos>
	 * into the FOLLOW_SET <f>, unless <f> is null.
	 */
	if (!f) return;

	/* Copy the count of FID's. */
	f->flen = fpos->count;

	/* Allocate memory to store the FID's. */
	f->follow = (FFID *)cgmalloc2(f->flen, sizeof(FID)+1, "efollow");
	
#ifdef	DEBUG
	printf("ID = %d LIT %c flen = %d\n", e->id, e->lit, f->flen);
#endif
	/* Step through the PSFID's in "fpos", last to first.  */
	for (id = fpos->last, p = f->follow; id > 0; id = q->previd, p++) {
		/* Get a pointer to the PSFID */
		q = PSFID_PTR(fpos, id);
		/* Copy the PSFID into a FID in the FOLLOW_SET array. */
		p->id = id;
		p->cost = q->cost;
		p->ins_cost = q->ins_cost;
#ifdef	DEBUG
	printf("FOLLOW ID = %d LIT %c\n", r->ptr[id]->id, r->ptr[id]->lit);
#endif
	}
	/* Check to make sure exactly "flen" FID's were copied. */
	if (p != f->follow + f->flen) {
		exitmsg2("%s: efollow length error: %d %d\n", p-f->follow, f->flen);
	}
}

/******************************************************************************

Routine:
	follow()

Syntax:

	static void follow(r, fpos, e, cost, bracket_no)

	RE_RE *r;	// Pointer to the regular expression we are matching
	POSITION_SET *fpos;  // Pointer to POSITION_SET to store following "lits".
	EXPR *e;	// Current expression
	short	cost;	// Accumulated mismatch cost, so far. 
	unsigned short bracket_no; // Angle bracket number associated with "e".

Description:

	follow() is called externally from efollow() to determine the set
	of "lit" expressions that can follow "lit" expression "e" in a match
	to text, taking approximate matching into account.

	To do the above, follow() calls first(), and first() and follow()
	may call themselves and each other recursively.  When called
	recursively from follow() or first(), "e" need not be a "lit",
	but can be any expression.

	When a "lit" expression is found that can follow "lit" "e" in a
	match is found, it is stored in "fpos".  Except for the
	End of Pattern expression, follow() itself does not find and store
	the "lits", first() does.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
follow(r, fpos, e, cost, bracket_no)
register RE_RE *r;
POSITION_SET *fpos;
register EXPR *e;
short cost;
unsigned short bracket_no;
{
	register EXPR *p, *e_next;
	register unsigned short	brack_no;

	/* Nothing follows the End of Pattern.  */
	if (e->type == EOP) {
		return;
	}

	/* To find the following "lits", we must step up to the parent. */
	p = e->parent;

	switch(p->type) {
	case EOP:
		/* We have reached the End of Pattern, which is a "lit".
		 * Add the EOP to the position set, and return.
		 */
		SET(fpos, p->id, cost, NO_INSERT);
		return;
	case Plus:
	case Star:
		/* We have reached a '+' or a '*', which means that
		 * whatever lies under the '+' or '*' can be matched
		 * repeatedly.  To handle the possibility that the
		 * text could repeat, we look for the first possible
		 * "lits" under the '+' or '*', by calling first().
		 * 
		 * In case we are doing an approximate match, we pass
		 * in "cost", the accumulated error cost so far, to
		 * make sure that we do not exceed the limit.
		 * Also, if the '+' or '*' is in the same angle bracket
		 * as the child "e", we pass in that bracket number
		 * to assure that insertions of extra characters
		 * are disallowed.  If either one is not in an angle
		 * bracket, or if they are in different angle brackets,
		 * we pass in 0 so that insertions will be allowed,
		 * if they don't go over the cost limit.
		 */
		brack_no = ((bracket_no == NO_INSERT || p->n_bracket == bracket_no) ? bracket_no : 0);
		(void)first(r, fpos, e, cost, brack_no);

		/* We will call follow() below, to handle matches
		 * where the matched text does not repeat. 
		 */
		break;
	case Cat:
		/* We have reached a concatenation.  Get the right
		 * (left) child of the parent, if we are matching
		 * forward (backward).
		 */
		e_next = (r->backward ? p->l : p->r);
		/* If "e_next" is not "e" itself, it contains the
		 * next "lit" in sequence.  Call first() to get the
		 * first possible "lits" in "e_next".  If "e_next" is not
		 * nullable (i.e., cannot match the null string), first()
		 * will return YES, and we return.
		 */
		if (e != e_next &&
		    first(r, fpos, e_next, cost, bracket_no)) {
			return;
		}
		/* Either "e" is the last child of the Cat ("e" == "e_next"),
		 * or "e" is the first child, but the last child can match
		 * the null string.  In either case, we will call
		 * follow() below to continue searching for more possible
		 * following "lits".
		 */
		break;
	}
	/* Now, continue by calling follow for the parent expression. */
	follow(r, fpos, p, cost, bracket_no);
}

/******************************************************************************

Routine:
	first()

Syntax:

	static BOOL first(r, fpos, e, cost, bracket_no)

	RE_RE *r;	// Pointer to the regular expression we are matching
	POSITION_SET *fpos;  // Pointer to POSITION_SET to store first "lits".
	EXPR *e;	// Current expression
	short	cost;	// Accumulated mismatch cost, so far. 
	unsigned short bracket_no; // Angle bracket number associated with
				   //  original expression (<e> at the
				   //  highest level call to first() or
				   //  follow()).

Description:

	first () is called to determine the first set of atomic "lit"
	expressions that can match a regular expression, beginning at
	expression <e>, taking approximate matching into account.

	To do the above, first() may call itself recursively.
	If approximate matching is being used, first() may also call follow(),
	which may call itself or first() recursively.

	When an atomic "lit" expression is found that can follow "e" in a match
	is found, it is stored in "fpos".

	EOP, although a "lit", is composite, so we do not store it.

Global Variables:

	None

Files:
	None

Return Status:

	YES	// "e" cannot exactly match an empty string.
		//   i.e., "e" is non-nullable.
	NO	// "e" can exactly match an empty string.
		//   i.e., "e" is nullable.

*******************************************************************************/

static BOOL
first(r, fpos, e, cost, bracket_no)
register RE_RE *r;
POSITION_SET *fpos;
register EXPR *e;
short cost;
unsigned short bracket_no;
{
	register EXPR *e_prev, *e_next;
	BOOL j;
	short new_cost;
	unsigned short ins_cost;

	switch(e->type) {
	case Literal:
	case Dot:
	case Charclass:
		if (bracket_no == NO_INSERT) {
			/* There is no adjacent "lit", i.e. <e> is the first
			 * character matched in the regular expression.
			 * No inserts before the initial character.
			 */
			ins_cost = NO_INSERT;
		} else if (bracket_no && e->n_bracket == bracket_no) {
			/* <e> and the adjacent "lit" are in the same
			 * angle brackets.  We cannot insert characters
			 * between them.
			 */
			ins_cost = NO_INSERT;
		} else {
			/* <e> and the adjacent "lit" are not in the same
			 * angle brackets.  We can insert characters
			 * between them, subject to the costs.
			 */
			ins_cost = r->cost_ins;
		}

		/* We have a "lit"  Store it in the POSITION_SET. */
		SET(fpos, e->id, cost, ins_cost);

		/* If "e" is not inside of angle brackets, it is possible
		 * that the character matching it is not in the text, and we
		 * still have a match due to it being considered a deletion
		 * error. If the additional cost of a deletion error does
		 * not put us over the mismatch cost limit, look for more
		 * following "lits" to put into the position set, albeit
		 * at a higher cost.
		 */
		if (!e->n_bracket &&
		    (new_cost = cost + r->cost_del) <= r->cost_limit) {
			follow(r, fpos, e, new_cost, bracket_no);
		}
		/* A "lit" is non-nullable. */
		j = YES;
		break;
	case Star:
	case Quest:
		/* Get the first "lit" inside the '*' or '?'.
		 * Since the '*' or '?' means zero instances
		 * will match, "e" is not non-nullable.
		 */
		(void)first(r, fpos, e->l, cost, bracket_no);
		j = NO;
		break;
	case Backref:
	case Group:
	case EOP:
	case Plus:
		/* Get the first "lit" inside.  "e" is non-nullable
		 * if and only if what is inside is non-nullable.
		 */
		j = first(r, fpos, e->l, cost, bracket_no);
		break;
	case Cat:
		/* Get the first lit inside the first expression
		 * in the concatenation.  If the first expression
		 * is not non-nullable, do the same for the second
		 * expression in the concatenation.  "e" is
		 * non-nullable if either expression is nullable.
		 */
		e_prev = (r->backward ? e->r : e->l);
		e_next = (r->backward ? e->l : e->r);
		j = first(r, fpos, e_prev, cost, bracket_no);
		j = (BOOL)(j || first(r, fpos, e_next, cost, bracket_no));
		break;
	case Alternate:
		/* Get the first lit inside the each expression
		 * in the alternation.  "e" is non-nullable if
		 * and only if both expressions are non-nullable.
		 */
		j = first(r, fpos, e->l, cost, bracket_no);
		j = (BOOL)(first(r, fpos, e->r, cost, bracket_no) && j);
		break;
	default:
		exitmsg("%s: bad type %d\n", e->type);
	}
	return(j);
}

/******************************************************************************

Routine:
	addstate()

Syntax:
	static STATE *addstate(r, ps)

	RE_RE *r;		// Pointer to the regular expression we are matching
	POSITION_SET *ps;	// Pointer to POSITION_SET containing the FID's
				//   that comprise the STATE to be added.

Description:

	addstate() is called when the new state of the Finite State Machine 
	is not already stored in memory.  addstate() stores the state.

Global Variables:

	None

Files:
	None

Return Status:

	A pointer to the new STATE is returned.

*******************************************************************************/

static STATE *
addstate(r, ps)
register RE_RE *r;
register POSITION_SET *ps;
{
	register ID j;
	register FID *p;
	register PSFID *q = 0;
	register STATE *s;

	/* Determine the next free slot in memory in which to store the STATE. */
	s = r->states + eg_getstate(r);

	/* The state transition table for the new state must be initialized
	 * to zeroes.
	 */
	(void)memset((char *)s->tab, 0, (int)sizeof(s->tab));

	/* The new state will not be a match to the entire regular expression,
	 * unless the End of Pattern (EOP) "lit" is in the position set.
	 */
	s->is_match = NO;

	/* Determine the location storage of the FID's for the state. */
	s->pos = eg_posalloc(r, ps->count);

	/* Set the count of FID's for the state. */
	s->npos = ps->count;

	/* Loop through the FID's in the POSITION_SET. */ 
	p = r->posbase + s->pos;
	for (j = ps->last; j > 0; p++, j = q->previd) {
		/* Get the FID, and store it. */
		q = PSFID_PTR(ps, j);
		p->id = j;
		p->cost = q->cost;
		p->ins_cost = q->ins_cost;
		/* If this FID is for the EOP "lit", (i.e. the root), then
		 * this state represents a match to the entire regular expression.
		 */
		if (p->id == r->root->id) {
			s->is_match = YES;
		}
	}
#ifdef	DEBUG
	if (TRACE(3)) {
		CHAR buf[2000];
		eg_spr_f(s->npos, s->pos+r->posbase, buf);
		PR("new state[%d] %s%s\n", s-r->states, buf, s->is_match?" is_match":"");
	}
#endif
	/* Return a pointer to the new state. */
	return(s);
}

#ifdef DEBUG
ppr(ps, p)
register POSITION_SET *ps;
register char *p;
{
	register ID n;
	register PSFID *q;

	if (ps->count < 1) {
		SPR(p, "{}");
		return;
	}
	*p++ = '{';
	for (n = ps->last; n > 0; n = q->previd) {
		q = PSFID_PTR(ps, n);
		SPR(p, "%d", n);
		p = strchr(p, 0);
	}
	p[-1] = '}';
}

void
eg_epr(e, res)
register EXPR *e;
char *res;
{
	CHAR r1[EPRINTSIZE], r2[EPRINTSIZE];
	register int i;

	if (e == 0) {
		SPR(res, "!0!");
		return;
	}
	switch(e->type) {
	case Dot:
	case Literal:
		eg_spr_c(e->b.flen, &e->b.follow, r1);
		eg_spr_c(e->f.flen, &e->f.follow, r1);
		SPR(res, "%c%s", e->lit, r1);
		break;
	case Charclass:
		*res++ = '[';
		for (i=0; i < NCHARS; i++) if (CCL_CHK((CHAR *)e->r, i)) {
			*res++ = i;
		}
		*res++ = ']';
		*res = 0;
		break;
	case Cat:
		eg_epr(e->l, r1);
		eg_epr(e->r, r2);
		SPR(res, "%s%s", r1, r2);
		break;
	case Alternate:
		eg_epr(e->l, r1);
		eg_epr(e->r, r2);
		SPR(res, "(%s|%s)", r1, r2);
		break;
	case Star:
		eg_epr(e->l, r1);
		SPR(res, "(%s)*", r1);
		break;
	case Plus:
		eg_epr(e->l, r1);
		SPR(res, "(%s)+", r1);
		break;
	case Quest:
		eg_epr(e->l, r1);
		SPR(res, "(%s)?", r1);
		break;
	case Group:
		eg_epr(e->l, r1);
		SPR(res, "(%s)", r1);
		break;
	case EOP:
		eg_epr(e->l, r1);
		SPR(res, "%s<EOP>", r1);
		break;
	case Backref:
		SPR(res, "\\%c", e->lit);
		break;
	default:
		SPR(res, "<undef type %d>", e->type);
		err(res);
		break;
	}
}

static void
eg_spr_f(c, p, buf)
int c;
FID *p;
register char *buf;
{
	if (c > 0) {
		*buf++ = '{';
		*buf = 0;
		while (--c > 0) {
			SPR(buf, "%d,", (p++)->id);
			buf = strchr(buf, 0);
		}
		SPR(buf, "%d}", p->id);
	} else {
		SPR(buf, "{}");
	}
}

static void
eg_spr_c(c, p, buf)
int c;
FID *p;
register char *buf;
{
	if (c > 0) {
		*buf++ = '{';
		*buf = 0;
		while (--c > 0) {
			SPR(buf, "%d,", (p++)->id);
			buf = strchr(buf, 0);
		}
		SPR(buf, "%d}", p->id);
	} else {
		SPR(buf, "{}");
	}
}
#endif
